/*----------------------------------------------------------------------------
 *-------------------------      M a R T E   O S      ------------------------
 *----------------------------------------------------------------------------
 *                                                             V1.57  Dic 2005
 *                                'm a l l o c'
 *
 *                                      C
 *
 * File 'malloc.c'                                   By Miguel Masmano Tello
 *                                                      and MAR.
 *
 * Implements two versions of malloc:
 *
 * - Two Levels Segregate Fit memory allocator (TLSF): This algorithm
 *   allows memory allocation and deallocation with bounded response
 *   times (Developed at the Universidad Politcnica de Valencia by
 *   Miguel Masmano Tello <mmasmano@disca.upv.es>)
 *
 * - Simple algorithm: allows allocation of memory blocks with not
 *   deallocation. It is useful for applications that does not use
 *   dynamic memory at all or uses it only at initialization time.
 *
 * ---------------------------------------------------------------------------
 *
 * Two Levels Segregate Fit memory allocator (TLSF)
 * Version 1.1
 *
 * Written by Miguel Masmano Tello <mmasmano@disca.upv.es>
 * Copyright (C) Dec, 2002 OCERA Consortium
 * Release under the terms of the GNU General Public License Version 2
 *
 * ----------------------------------------------------------------------
 *  Copyright (C) 2004   Universidad de Cantabria, SPAIN
 *
 *  MaRTE OS web page: http://marte.unican.es
 *  Contact Addresses: Mario Aldea Rivas          aldeam@unican.es
 *                     Michael González Harbour      mgh@unican.es
 *
 * MaRTE OS  is free software; you can  redistribute it and/or  modify it
 * under the terms of the GNU General Public License  as published by the
 * Free Software Foundation;  either  version 2, or (at  your option) any
 * later version.
 *
 * MaRTE OS  is distributed  in the  hope  that  it will be   useful, but
 * WITHOUT  ANY  WARRANTY;     without  even the   implied   warranty  of
 * MERCHANTABILITY  or  FITNESS FOR A  PARTICULAR PURPOSE.    See the GNU
 * General Public License for more details.
 *
 * You should have received  a  copy of  the  GNU General Public  License
 * distributed with MaRTE  OS;  see file COPYING.   If not,  write to the
 * Free Software  Foundation,  59 Temple Place  -  Suite 330,  Boston, MA
 * 02111-1307, USA.
 *
 * As a  special exception, if you  link this  unit  with other  files to
 * produce an   executable,   this unit  does  not  by  itself cause  the
 * resulting executable to be covered by the  GNU General Public License.
 * This exception does  not however invalidate  any other reasons why the
 * executable file might be covered by the GNU Public License.
 *
 *---------------------------------------------------------------------------*/
#include <sys/marte_configuration_parameters.h>

/*
 * Pool of dynamic memory)
 */
char dynamic_memory_pool [DYNAMIC_MEMORY_POOL_SIZE_IN_K * 1024];


#ifdef USE_TLSF_MEMORY_ALLOCATOR
/*
 * Two Levels Segregate Fit memory allocator (TLSF)
 * Version 1.1
 *
 * Written by Miguel Masmano Tello <mmasmano@disca.upv.es>
 * Copyright (C) Dec, 2002 OCERA Consortium
 * Release under the terms of the GNU General Public License Version 2
 *
 */
#include <stdio.h>
#include <string.h>
#include "tlsf.h"
#include "tlsf_mapping.h"

/*
 * TLSF structure (pointer to the pool of dynamic memory)
 */
// << MaRTE 
TLSF_t *ptr_TLSF;
// MaRTE >>

/*
 * header_overhead has size of blocks_header - 2 * ptr to next free block
 */
// << MaRTE
//static __s32 beg_header_overhead = 0;
__s32 beg_header_overhead = 0;
// MaRTE >>


/* 
 * init_dynamic_memory
 *
 * max_sl_log2_index defines  the maximum  second index which  will be
 * used by TLSF.
 *
 * max_sl_log2_index allows  to the user to tune  the maximum internal
 * fragmentation, but  a high  max_sl_log2_index value will  cause big
 * TLSF structure.
 *
 * max_sl_log2_index  max. internal fragmentation (approximately)
 * -----------------  -------------------------------------------
 *     1                             25 %
 *     2                           12.5 %
 *     3                           6.25 %
 *     4                          3.125 %
 *     5                          1.563 %
 *
 * Max_size is in Kbytes, On success this function returns the free size
 */
int init_dynamic_memory ()
{
  __s32 n, free_mem = 0, total_size;
  block_header_t *initial_block_ptr;
  __s32 size_fl_sl_array, i, fl, sl;
  // << MaRTE
  // TLSF_t *ptr_TLSF;
  int max_size = DYNAMIC_MEMORY_POOL_SIZE_IN_K;
  int max_sl_log2_index = TLSF_MAX_SL_LOG2_INDEX; 
  char *block_ptr  = dynamic_memory_pool;
  // MaRTE >>
  
  if (!(max_size > 0)) {
    PRINT_DBG_C ("ERROR: size must be > 0\n");
    return -1;
  }
  
  if (max_sl_log2_index > 5 || max_sl_log2_index < 1) {
    PRINT_DBG_C ("ERROR max_sl_log2_index must be >=1 or <= 5\n");
    return -1;
  }
  
  if (max_sl_log2_index <= 0) {
    PRINT_DBG_C ("ERROR: max_sl_log2_index must be >= 0\n");
    return -1;
  }

  if ((((__u32) block_ptr >> LOG2_TLSF_WORD_SIZE) << LOG2_TLSF_WORD_SIZE) 
      != (__u32) block_ptr) {
    PRINT_DBG_C ("ERROR block_ptr must be aligned\n");
    return -1;
  }
  
  memset ((char *) block_ptr, 0x00, max_size * 1024);

  // << MaRTE
  // INIT_THREAD_MUTEX();
  // MaRTE >>
  ptr_TLSF = (TLSF_t *) block_ptr;
  
  ptr_TLSF -> magic_number = MAGIC_NUMBER;
  
  /* Total size of the block, TLSF_struct + free_memory */
  
  total_size = BYTES2TLSF_WORDS(max_size * 1024);

  ptr_TLSF -> max_sl_log2_index = max_sl_log2_index;
  ptr_TLSF -> max_sl_index = (1 << ptr_TLSF -> max_sl_log2_index);
  
  size_fl_sl_array = BYTES2TLSF_WORDS((__s32) sizeof (fl_array_t) +
    ((__s32) sizeof (block_header_t *) * (__s32) ptr_TLSF -> max_sl_index));

  free_mem = total_size - BYTES2TLSF_WORDS(sizeof (TLSF_t)) - size_fl_sl_array;
    
  n = MIN_LOG2_SIZE + 1;
  
  while ((int) TLSF_WORDS2BYTES(free_mem) > 
	 (1 << (n + LOG2_TLSF_WORD_SIZE)) && n < MAX_FL_INDEX) {
    n ++;
    free_mem -= size_fl_sl_array;
  }
  
  if (free_mem < 0) return -1;

  // max_fl_index never will be greater than 32 (4 Gbytes)
  
  ptr_TLSF -> max_fl_index = n;
  ptr_TLSF -> max_fl_pow2_index = (1 << ptr_TLSF -> max_fl_index);
  
  n -= MIN_LOG2_SIZE;
  
  /* max_fl_index will never be greater than MAX_FL_INDEX */
  if (ptr_TLSF -> max_fl_index < 0 || MAX_FL_INDEX < 0) return -1;
  
  ptr_TLSF -> fl_array = ( fl_array_t *) 
    ((__u32) &(ptr_TLSF -> fl_array) 
     + (__u32) sizeof (ptr_TLSF -> fl_array));
  
  for (i = 0 ; i < n; i ++)
    ptr_TLSF -> fl_array [i] .sl_array = (block_header_t **) 
      (((__s32) ptr_TLSF -> fl_array + ((__s32) sizeof (fl_array_t) * n)) +
       ((__s32) sizeof (block_header_t *) *
	(__s32) ptr_TLSF -> max_sl_index * i));
  
  initial_block_ptr = (block_header_t *) 
    ((__u32) ptr_TLSF -> fl_array + 
     (TLSF_WORDS2BYTES(size_fl_sl_array) * n));
  
  ptr_TLSF -> first_bh = initial_block_ptr;
  
  beg_header_overhead = BYTES2TLSF_WORDS((int) initial_block_ptr -> ptr.buffer
					 - (int) initial_block_ptr);

  ptr_TLSF -> bitmapFL = 0;
 
  
  initial_block_ptr -> size =
    free_mem - beg_header_overhead;

  SET_LAST_BLOCK (initial_block_ptr);

  initial_block_ptr -> ptr.free_ptr.prev = NULL;
  initial_block_ptr -> ptr.free_ptr.next = NULL;
  initial_block_ptr -> prev_phys_block = NULL;
  
  if (TLSF_WORDS2BYTES(GET_BLOCK_SIZE(initial_block_ptr)) 
      <= MIN_SIZE) {
    return -1;
  } else {
    mapping_function (GET_BLOCK_SIZE(initial_block_ptr), &fl, &sl, 
		      ptr_TLSF);
    fl -= MIN_LOG2_SIZE;
  }

  ptr_TLSF -> fl_array [fl].sl_array [sl] = initial_block_ptr;
  TLSF__set_bit (sl, ptr_TLSF -> fl_array[fl].bitmapSL);
  TLSF__set_bit (fl, ptr_TLSF -> bitmapFL);

  printc ("TLSF Dynamic Memory Pool Initialized"
	  " (%d bytes free, starts at %p)\n",
	  TLSF_WORDS2BYTES(GET_BLOCK_SIZE(initial_block_ptr)),
	  initial_block_ptr);
  
  return TLSF_WORDS2BYTES(GET_BLOCK_SIZE(initial_block_ptr));
}


#else // USE_TLSF_MEMORY_ALLOCATOR

/*
 * Symple dynamic memory algorithm
 */
extern void simple_dynamic_memory__initialize (char * pool);
extern void simple_dynamic_memory___elabb ();

int init_dynamic_memory ()
{
  simple_dynamic_memory___elabb ();
  simple_dynamic_memory__initialize (dynamic_memory_pool);
  return DYNAMIC_MEMORY_POOL_SIZE_IN_K * 1024;
}

#endif // USE_TLSF_MEMORY_ALLOCATOR

