------------------------------------------------------------------------------
--  ------------------         M a R T E   O S         -------------------  --
------------------------------------------------------------------------------
--                                                             V1.57  Dic 2005
--
--                    'K e r n e l . S c h e d u l e r'
--
--                                 Body
--
--
--  File 'k-scheduler.adb'                                            By MAR.
--
--
--  Scheduler operations.
--
--  ----------------------------------------------------------------------
--   Copyright (C) 2004   Universidad de Cantabria, SPAIN
--
--   MaRTE OS web page: http://marte.unican.es
--   Contact Addresses: Mario Aldea Rivas          aldeam@unican.es
--                      Michael González Harbour      mgh@unican.es
--
--  MaRTE OS  is free software; you can  redistribute it and/or  modify it
--  under the terms of the GNU General Public License  as published by the
--  Free Software Foundation;  either  version 2, or (at  your option) any
--  later version.
--
--  MaRTE OS  is distributed  in the  hope  that  it will be   useful, but
--  WITHOUT  ANY  WARRANTY;     without  even the   implied   warranty  of
--  MERCHANTABILITY  or  FITNESS FOR A  PARTICULAR PURPOSE.    See the GNU
--  General Public License for more details.
--
--  You should have received  a  copy of  the  GNU General Public  License
--  distributed with MaRTE  OS;  see file COPYING.   If not,  write to the
--  Free Software  Foundation,  59 Temple Place  -  Suite 330,  Boston, MA
--  02111-1307, USA.
--
--  As a  special exception, if you  link this  unit  with other  files to
--  produce an   executable,   this unit  does  not  by  itself cause  the
--  resulting executable to be covered by the  GNU General Public License.
--  This exception does  not however invalidate  any other reasons why the
--  executable file might be covered by the GNU Public License.
--
-------------------------------------------------------------------------------

with Hardware_Interface;
pragma Elaborate_All (Hardware_Interface);

with Basic_Integer_Types; use Basic_Integer_Types;
pragma Elaborate_All (Basic_Integer_Types);
with Option_Sets;
pragma Elaborate_All (Option_Sets);

with Kernel;
pragma Elaborate (Kernel);
with Kernel.Application_Scheduler;
pragma Elaborate (Kernel.Application_Scheduler);
with Kernel.Application_Scheduling_Data;
with Kernel.Timed_Events_And_Timer;
pragma Elaborate_All (Kernel.Timed_Events_And_Timer);
with Kernel.Timed_Events_Queue;
pragma Elaborate_All (Kernel.Timed_Events_Queue);
with Kernel.Task_Suspension_Timed_Events;
pragma Elaborate_All (Kernel.Task_Suspension_Timed_Events);
with Kernel.Replenishment_TE;
pragma Elaborate_All (Kernel.Replenishment_TE);
with Kernel.Condition_Variables.Internals;
with Kernel.Mutexes.Internals;
with Kernel.Semaphores.Internals;
with Kernel.Pool_TCBs; -- to use 'Main_Task_TBC'
with Kernel.Tasks_Operations.Internals;
--  pragma Elaborate (Kernel.Tasks_Operations.Internals);
with Kernel.Tasks_Operations.Initialize_TCBs;
pragma Elaborate (Kernel.Tasks_Operations.Initialize_TCBs);
with Kernel.Timers;
with Kernel.Timers.Internals;
with Kernel_Console;
pragma Elaborate_All (Kernel_Console);
with Kernel.Hardware_Interrupts;
pragma Elaborate (Kernel.Hardware_Interrupts);

--  Debug
with Kernel.Scheduler.Debug;
--  with Kernel.Scheduler.CDbg; -- {CDBG}
--  pragma Elaborate_All (Kernel.Scheduler.CDbg);


package body Kernel.Scheduler is

   package ML                 renames Kernel.Tasks_Map_Lists;
   package TSTE               renames Kernel.Task_Suspension_Timed_Events;
   package APPSCHD            renames Kernel.Application_Scheduler;
   package APPSCHD_DATA       renames Kernel.Application_Scheduling_Data;
   package TE_T               renames Kernel.Timed_Events_And_Timer;
   package TEQ                renames Kernel.Timed_Events_Queue;
   package TOI                renames Kernel.Tasks_Operations.Internals;

   use HWI,
     Kernel.Tasks_Lists,
     Kernel.Timed_Events_Lists,
     Kernel.CPU_Time_Timed_Events_Lists,
     Kernel.CPU_Time_Timed_Events_Lists_Order;

   function UC (Base : Kernel.AppScheduler_Data_Base_Ac)
                return APPSCHD_DATA.AppScheduler_Data_Ac
     renames APPSCHD_DATA.UC_To_AppScheduler_Data_Ac;
   pragma Inline (UC);
   function UC (Base : Kernel.Sched_App_Data_Base_Ac)
                return APPSCHD_DATA.Sched_App_Data_Ac
     renames APPSCHD_DATA.UC_To_Sched_App_Data_Ac;
   pragma Inline (UC);


   --------------------------------------
   -- Round_Robin_Interval_Time_Period --
   --------------------------------------
   Round_Robin_Interval_Time_Period : constant HWTime :=
     HWI.Duration_To_HWTime
        (Duration (Configuration_Parameters.Round_Robin_Interval_Time_Period));

   ------------------
   -- Running_Task --
   ------------------
   Running_Task : Task_Id;

   -----------------
   -- Ready Queue --
   -----------------
   Ready_Queue : ML.Map_List;

   -----------------------------
   -- Interrupt_Nesting_Count --
   -----------------------------
   --
   --  Hardware interrupt nesting counter (modified in the interrupt
   --  handlers in 'x86/boot/base_irq_inittab.S'), used to ensure that the
   --  software interrupt handler ('Do_Scheduling') isn't called until all
   --  outstanding hardware interrupts have been processed. In addition,
   --  this byte also acts as the software interrupt pending flag: if the
   --  high bit is set, no software interrupt is pending; if the high bit
   --  is clear, a software interrupt is pending. This design allows the
   --  interrupt path to decide whether to call the software interrupt
   --  handler simply by testing the word for zero.
   Interrupt_Nesting_Count : Unsigned_8;
   pragma Import (C, Interrupt_Nesting_Count, "base_irq_nest");
   --  'base_irq_nest' is defined in 'x86/boot/base_irq.c'.
   Outside_Of_Interrupts : constant := 16#80#;
   Invoke_Do_Scheduling_Mask : constant := 16#7F#; -- not 16#80#;

   ----------
   -- Self --
   ----------
   function Self return Task_Id is
   begin
      pragma Debug (Debug.Assert_Task_Is_Ok (Running_Task));
      return Running_Task;
   end Self;

   ------------------------------------------------
   -- Dequeue_Task_Ending_Reenqueuement_Deferred --
   ------------------------------------------------
   procedure Dequeue_Task_Ending_Reenqueuement_Deferred (T : in Task_Id);
   procedure Dequeue_Task_Ending_Reenqueuement_Deferred (T : in Task_Id) is
      Real_Active_Prio : Task_Priority;
   begin
      pragma Assert (Task_OK (T));
      pragma Assert (T.Status /= READY);
      pragma Assert (T.Reenqueuement_Deferred);
      Real_Active_Prio := T.Active_Prio;
      T.Active_Prio    := T.Old_Active_Prio;
      ML.Dequeue_First_In_Priority (T, Ready_Queue);
      T.Active_Prio    := Real_Active_Prio;
      T.Reenqueuement_Deferred := False;
   end Dequeue_Task_Ending_Reenqueuement_Deferred;
   pragma Inline (Dequeue_Task_Ending_Reenqueuement_Deferred);

   -----------------------------------------
   -- Schedule_SS_Replenishment_Operation --
   -----------------------------------------
   procedure Schedule_SS_Replenishment_Operation
     (Now : in HWTime; Limit_Reached : in Boolean := False);
   procedure Schedule_SS_Replenishment_Operation
     (Now : in HWTime; Limit_Reached : in Boolean := False) is
      Repl_Event : Replenishment_TE.Replenishment_Timed_Event_Ac;
   begin
      Running_Task.Number_Of_Pending_Replenishments :=
        Running_Task.Number_Of_Pending_Replenishments + 1;

      Repl_Event := Replenishment_TE.Request;

      Repl_Event.SS_Task := Running_Task;
      Repl_Event.T := Running_Task.Activation_Time
        + Running_Task.Sched_SS_Repl_Period;
      if Limit_Reached or else Running_Task.Scheduler_Timed_Event.T < Now then
         Repl_Event.Replenish_Amount :=
           Running_Task.Capacity - Running_Task.Replenish_Amount_Pending;
         Running_Task.Replenish_Amount_Pending := Running_Task.Capacity;
      else
         Repl_Event.Replenish_Amount :=
           Running_Task.Capacity - (Running_Task.Scheduler_Timed_Event.T - Now)
           - Running_Task.Replenish_Amount_Pending;
         Running_Task.Replenish_Amount_Pending :=
           Running_Task.Replenish_Amount_Pending + Repl_Event.Replenish_Amount;
      end if;

      --  pragma Debug -- {CDBG}
      --    (CDbg.Schedule_SS_Replenishment_Operation -- {CDBG}
      --     (Running_Task.Activation_Time, Repl_Event.T, -- {CDBG}
      --      Repl_Event.Replenish_Amount, -- {CDBG}
      --      Running_Task.Replenish_Amount_Pending)); -- {CDBG}

      TE_T.Enqueue_Standard_Event_Without_Updating_Timer
        (Timed_Event_Ac (Repl_Event));
   end Schedule_SS_Replenishment_Operation;
   pragma Inline (Schedule_SS_Replenishment_Operation);

   --------------------------------------------------
   -- Reenqueue_Task_Ending_Reenqueuement_Deferred --
   --------------------------------------------------
   procedure Reenqueue_Task_Ending_Reenqueuement_Deferred (T : in Task_Id);
   procedure Reenqueue_Task_Ending_Reenqueuement_Deferred (T : in Task_Id) is
      Real_Active_Prio : Task_Priority;
   begin
      pragma Assert (Task_OK (T));
      pragma Assert (T.Status = READY);
      pragma Assert (T.Reenqueuement_Deferred);
      Real_Active_Prio := T.Active_Prio;
      T.Active_Prio    := T.Old_Active_Prio;
      ML.Dequeue_First_In_Priority (T, Ready_Queue);
      T.Active_Prio    := Real_Active_Prio;
      ML.Enqueue_First_In_Priority (T, Ready_Queue);
      --  I can use 'Enqueue_First_In_Priority' instead of
      --  'Enqueue_In_Order' because I know there isn't another task in the
      --  'Running_Task' active priority queue, otherwise the reenqueuement
      --  deferred would has been already ended.
      T.Reenqueuement_Deferred := False;
   end Reenqueue_Task_Ending_Reenqueuement_Deferred;
   pragma Inline (Reenqueue_Task_Ending_Reenqueuement_Deferred);

   ------------------------
   -- Register_Main_Task --
   ------------------------
   procedure Register_Main_Task (Main : in out Task_Id);
   procedure Register_Main_Task (Main : in out Task_Id) is
   begin
      Running_Task := Main;

      --  Put Main Task ready
      Sched_Fifo_Task_Gets_Ready (Main);
   end Register_Main_Task;

   ------------------------
   -- Register_Idle_Task --
   ------------------------
   procedure Register_Idle_Task (Idle : in Task_Id);
   procedure Register_Idle_Task (Idle : in Task_Id) is
   begin
      --  Idle Task in Ready Queue
      Idle.Status := READY;
      ML.Set_Last_Element (Idle, Ready_Queue);
   end Register_Idle_Task;


   ----------------------------------------------------------------------------
   -- Functions and procedures of the interface -------------------------------
   ----------------------------------------------------------------------------

   ----------------------
   -- New_Task_Created --
   ----------------------
   procedure New_Task_Created (T : in Task_Id) is
   begin
      pragma Assert (Task_OK (T, NEW_APPSCHED) and then
                     (T.Status = BLOCKED or T.Status = APPSCHED_SUSPENDED),
                     "New_Task_Created: invalid Task");

      case T.Sched_Policy is
         when SCHED_FIFO | SCHED_RR | SCHED_SPORADIC | SCHED_OTHER =>
            --  Put the task Ready
            Task_Gets_Ready (T);
            if T.Sched_Policy = SCHED_SPORADIC then
               --  Enqueue scheduler event in the CPU timed events queue. This
               --  event is set to expire when the capacity of the task
               --  finishes.
               T.Scheduler_Timed_Event.CPU_Time :=
                 T.Used_CPU_Time + T.Capacity;
               CPU_Time_Timed_Events_Lists_Order.Enqueue_In_Order
                 (T.Scheduler_Timed_Event'Access, T.CPU_Time_TEs_Q);

               --  Of course the task starts with capacity remaining
               T.Capacity_Exhausted := False;
            end if;

         when SCHED_APP =>
            --  Suspend father task waiting for appscheduler response
            --  (only if the NEW event is unmasked for the appscheduler)
            --  and send event to the appscheduler.
            APPSCHD.Send_Event_From_Task
              (Event_Code => APPSCHD_DATA.APPSCHED_NEW,
               ST         => UC (T.Sched_App).Scheduler,
               T          => T);
      end case;
   end New_Task_Created;

   -------------------------------
   -- Running_Task_Gets_Blocked --
   -------------------------------
   procedure Running_Task_Gets_Blocked (With_Status : in Task_Status) is
   begin
      pragma Assert (Task_OK (Running_Task, TERMINATED));
      pragma Assert (Running_Task.Status = READY);
      pragma Assert (not (Running_Task.Task_Type = SIGNAL_HANDLER_TASK and
                          With_Status /= SIG_HANDLER_WAIT_TO_DELIVER));
      pragma Assert (Interrupt_Nesting_Count = Outside_Of_Interrupts);
      --  pragma Debug (CDbg.Task_Gets_Blocked (Running_Task)); -- {CDBG}

      Running_Task.Status := With_Status;

      if Running_Task.Reenqueuement_Deferred then
         Dequeue_Task_Ending_Reenqueuement_Deferred (Running_Task);
      else
         ML.Dequeue_Head (Running_Task, Ready_Queue);
      end if;

      case Running_Task.Sched_Policy is
         when SCHED_FIFO | SCHED_OTHER =>
            --  Is 'Running_Task' an appscheduler ?
            if (Running_Task.AppScheduler /= null  and then
                not APPSCHD_DATA.Sched_App_Data_Lists.Is_Empty
                (UC (Running_Task.AppScheduler).Suspension_Deferred_Tasks))
            then
               --  When an appscheduler gets blocked every task in its
               --  'Suspension_Deferred_Tasks' list has to be suspended as
               --  well.
               declare
                  package SCHDAPP_DATA_LST
                    renames APPSCHD_DATA.Sched_App_Data_Lists;
                  use SCHDAPP_DATA_LST;
                  Suspension_Deferred_Tasks : SCHDAPP_DATA_LST.List renames
                    UC (Running_Task.AppScheduler).Suspension_Deferred_Tasks;
               begin
                  loop
                     pragma Assert
                       (Task_OK (Head (Suspension_Deferred_Tasks).Owner));
                     Non_Running_Task_Gets_Blocked
                       (Head (Suspension_Deferred_Tasks).Owner,
                        APPSCHED_SUSPENDED);

                     Dequeue_Head (Suspension_Deferred_Tasks);
                     exit when Is_Empty (Suspension_Deferred_Tasks);
                  end loop;
               end;
            end if;
         when SCHED_RR =>
            CPU_Time_Timed_Events_Lists.Dequeue
              (Running_Task.Scheduler_Timed_Event'Access,
               Running_Task.CPU_Time_TEs_Q);
         when SCHED_SPORADIC =>
            if not Running_Task.Capacity_Exhausted then
               --  Limit of replenishments reached ?
               if (Running_Task.Number_Of_Pending_Replenishments >=
                   Running_Task.Sched_SS_Max_Repl - 1) then
                  --  Schedule "limit reached" Replenishment Operation
                  Schedule_SS_Replenishment_Operation (Now => Get_HWTime,
                                                       Limit_Reached => True);
                  --  Dequeue the excution CPU time limit event.
                  CPU_Time_Timed_Events_Lists.Dequeue
                    (Running_Task.Scheduler_Timed_Event'Access,
                     Running_Task.CPU_Time_TEs_Q);
                  --  Reduce priority
                  TOI.Set_Base_Prio (Running_Task,
                                     Running_Task.Sched_SS_Low_Priority);
                  Running_Task.Capacity_Exhausted := True;
               else
                  --  Schedule normal Replenishment Operation
                  Schedule_SS_Replenishment_Operation (Now => Get_HWTime);
               end if;
            end if;
         when SCHED_APP =>
            if (With_Status /= APPSCHED_SUSPENDED             and
                With_Status /= WAITING_TO_LOCK_APPSCHED_MUTEX and
                With_Status /= TRYING_TO_LOCK_APPSCHED_MUTEX  and
                With_Status /= WAITING_APPSCHEDULER_RESPONSE) then
               APPSCHD.Send_Event_From_Task
                 (Event_Code => APPSCHD_DATA.APPSCHED_BLOCK,
                  ST         => UC (Running_Task.Sched_App).Scheduler,
                  T          => Running_Task);
            end if;
            --  When the suspend reason is 'WAITING_TO_LOCK_APPSCHED_MUTEX'
            --  or 'TRYING_TO_LOCK_APPSCHED_MUTEX' the events are sent from
            --  functions in 'Kernel.Mutexes'.
      end case;
   end Running_Task_Gets_Blocked;

   -----------------------------------
   -- Non_Running_Task_Gets_Blocked --
   -----------------------------------
   --  By now only used for 'SCHED_APP' tasks. It do NOT send a BLOCK
   --  event to the appscheduler.
   procedure Non_Running_Task_Gets_Blocked (T           : in Task_Id;
                                            With_Status : in Task_Status) is
   begin
      pragma Assert (Task_OK (T, TERMINATED));
      pragma Assert (T.Sched_Policy = SCHED_APP or
                     With_Status = WAITING_POLICY_CHANGE_TO_SCHED_APP);
      pragma Assert (T.Status = READY);
      --  pragma Debug (CDbg.Task_Gets_Blocked (T)); -- {CDBG}

      T.Status := With_Status;

      if T.Reenqueuement_Deferred then
         Dequeue_Task_Ending_Reenqueuement_Deferred (T);
      else
         ML.Dequeue (T, Ready_Queue);
      end if;
   end Non_Running_Task_Gets_Blocked;

   --------------------------------
   -- Sched_Fifo_Task_Gets_Ready --
   --------------------------------
   procedure Sched_Fifo_Task_Gets_Ready (T : in Task_Id) is
   begin
      pragma Assert (Kernel.Task_OK (T),
                     "Sched_Fifo_Task_Gets_Ready: invalid Task");
      pragma Assert (T.Status /= READY,
                     "Sched_Fifo_Task_Gets_Ready: already ready Task");
      pragma Assert (not T.Reenqueuement_Deferred,
                     "Sched_Fifo_Task_Gets_Ready: Reenque. Deferred Task");
      pragma Assert
        (T.Sched_Policy = SCHED_FIFO or T.Sched_Policy = SCHED_OTHER,
           "Sched_Fifo_Task_Gets_Ready: no Sched FIFO or OTHER Task");
      --  pragma Debug (CDbg.Task_Gets_Ready (T)); -- {CDBG}

      T.Status := READY;
      ML.Enqueue_In_Order (T, Ready_Queue);
   end Sched_Fifo_Task_Gets_Ready;

   ---------------------
   -- Task_Gets_Ready --
   ---------------------
   procedure Task_Gets_Ready (T : in Task_Id) is
   begin
      pragma Assert (Kernel.Task_OK (T),
                     "Task_Gets_Ready: invalid Task");
      pragma Assert (T.Status /= READY,
                     "Task_Gets_Ready: already ready Task");
      pragma Assert (not T.Reenqueuement_Deferred,
                     "Task_Gets_Ready: Reenqueuement Deferred Task");

      case T.Sched_Policy is
         when SCHED_FIFO | SCHED_OTHER =>
            null;

         when SCHED_RR =>
            T.Scheduler_Timed_Event.CPU_Time := T.Used_CPU_Time + T.Capacity;
            CPU_Time_Timed_Events_Lists_Order.Enqueue_In_Order
              (T.Scheduler_Timed_Event'Access, T.CPU_Time_TEs_Q);

         when SCHED_SPORADIC =>
            if not T.Capacity_Exhausted then
               T.Activation_Time := Get_HWTime;
            end if;

         when SCHED_APP =>
            if (T.Status /= APPSCHED_SUSPENDED and
                T.Status /= WAITING_TO_LOCK_APPSCHED_MUTEX and
                T.Status /= TRYING_TO_LOCK_APPSCHED_MUTEX and
                T.Status /= WAITING_APPSCHEDULER_RESPONSE) then
               --  Only send notification to the application scheduler in
               --  the case the task activation hasn't been ordered by it.
               T.Status := APPSCHED_SUSPENDED;
               APPSCHD.Send_Notification_Or_Ready
                 (Event_Code => APPSCHD_DATA.APPSCHED_READY,
                  ST         => UC (T.Sched_App).Scheduler,
                  T          => T);
            else
               --  pragma Debug (CDbg.Task_Gets_Ready (T)); -- {CDBG}
               T.Status := READY;
               ML.Enqueue_In_Order (T, Ready_Queue);
            end if;
      end case;

      if T.Sched_Policy /= SCHED_APP then
         --  pragma Debug (CDbg.Task_Gets_Ready (T)); -- {CDBG}
         T.Status := READY;
         if T.Task_Type = SIGNAL_HANDLER_TASK then
            ML.Enqueue_First_In_Priority (T, Ready_Queue);
         else
            ML.Enqueue_In_Order (T, Ready_Queue);
         end if;
      end if;
   end Task_Gets_Ready;

   ---------------------------------------
   -- Ready_Task_Raises_Active_Priority --
   ---------------------------------------
   procedure Ready_Task_Raises_Active_Priority (T        : in Task_Id;
                                                New_Prio : in Task_Priority) is
   begin
      pragma Assert (Task_OK (T),
                     "Ready_Task_Raises_Active_Priority: invalid Task");
      pragma Assert (T.Status = READY,
                     "Ready_Task_Raises_Active_Priority: non-ready Task");
      --  pragma Debug (CDbg.Ready_Task_Raises_Active_Priority  -- {CDBG}
      --                  (T, New_Prio)); -- {CDBG}

      if T.Reenqueuement_Deferred then
         T.Active_Prio := New_Prio;
         Reenqueue_Task_Ending_Reenqueuement_Deferred (T);
      else
         ML.Dequeue (T, Ready_Queue);
         T.Active_Prio := New_Prio;
         ML.Enqueue_In_Order (T, Ready_Queue);
      end if;
   end Ready_Task_Raises_Active_Priority;

   ----------------------------------------
   -- Ready_Task_Reduces_Active_Priority --
   ----------------------------------------
   procedure Ready_Task_Reduces_Active_Priority
     (T                  : in Task_Id;
      New_Prio           : in Task_Priority;
      Head_Of_New_Prio_Q : in Boolean) is
   begin
      pragma Assert (Task_OK (T),
                     "Task_Reduces_Active_Priority: invalid Task");
      pragma Assert (T.Status = READY,
                     "Ready_Task_Reduces_Active_Priority: non-ready Task");
      --  pragma Debug (CDbg.Ready_Task_Reduces_Active_Priority -- {CDBG}
      --                  (T, New_Prio)); -- {CDBG}

      if T.Reenqueuement_Deferred then
         T.Active_Prio := New_Prio;
      else
         ML.Dequeue (T, Ready_Queue);
         T.Active_Prio := New_Prio;
         if Head_Of_New_Prio_Q then
            ML.Enqueue_First_In_Priority (T, Ready_Queue);
         else
            ML.Enqueue_In_Order (T, Ready_Queue);
         end if;
      end if;
   end Ready_Task_Reduces_Active_Priority;

   -------------------------
   -- Task_Changes_Policy --
   -------------------------
   procedure Task_Changes_Policy (T      : in Task_Id;
                                  Policy : in TO.Scheduling_Policies) is
      Now : constant HWI.HWTime := HWI.Get_HWTime;
   begin
      pragma Assert (Task_OK (T),
                     "Task_Changes_Policy: invalid Task");
      if T.Sched_Policy /= Policy then
         --  Actions due to former policy
         case T.Sched_Policy is
            when SCHED_FIFO | SCHED_OTHER =>
               null;
            when SCHED_RR | SCHED_SPORADIC =>
               if Is_In_The_List (T.Scheduler_Timed_Event'Access,
                                  T.CPU_Time_TEs_Q) then
                  Dequeue (T.Scheduler_Timed_Event'Access, T.CPU_Time_TEs_Q);
               end if;
            when SCHED_APP =>
               APPSCHD.Send_Event_From_Task
                 (Event_Code => APPSCHD_DATA.APPSCHED_TERMINATE,
                  ST         => UC (T.Sched_App).Scheduler,
                  T          => T);
               UC (T.Sched_App).Scheduler := null;
         end case;
         --  Actions due to new policy
         case Policy is
            when SCHED_FIFO | SCHED_OTHER =>
               null;
            when SCHED_RR | SCHED_SPORADIC =>
               --  (Notice that changing SCHED_SPORADIC task's parameters is
               --  only allowed when there are no pending replenishment
               --  operations)
               if T = Running_Task then
                  T.Scheduler_Timed_Event.CPU_Time :=
                    T.Used_CPU_Time + T.Capacity;
                  T.Scheduler_Timed_Event.T :=
                    T.Running_Starting_Time + T.Capacity;
                  Enqueue_In_Order (T.Scheduler_Timed_Event'Access,
                                    T.CPU_Time_TEs_Q);
                  TE_T.Update_Timer_Due_To_New_CPU_Event
                    (T.Scheduler_Timed_Event'Access, Now);
               else
                  T.Scheduler_Timed_Event.CPU_Time :=
                    T.Used_CPU_Time + T.Capacity;
                  Enqueue_In_Order (T.Scheduler_Timed_Event'Access,
                                    T.CPU_Time_TEs_Q);
               end if;
            when SCHED_APP =>
               null;
         end case;
      end if;
   end Task_Changes_Policy;

   --------------------------------------
   -- New_Sched_Param_In_AppSched_Task --
   --------------------------------------
   procedure New_Sched_Param_In_AppSched_Task (T : in Task_Id) is
   begin
      pragma Assert (Task_OK (T),
                     "New_Sched_Param_In_AppSched_Task: invalid Task");
      pragma Assert (T.Sched_Policy = SCHED_APP,
                       "New_Sched_Param_In_AppSched_Task: no SCHED_APP Task");

      if UC (T.Sched_App).Scheduler /= Running_Task then
         --  Send the event to the scheduler only if the parameters
         --  are not changed from the appscheduler its self
         APPSCHD.Send_Event_From_Task
           (Event_Code => APPSCHD_DATA.APPSCHED_CHANGE_SCHED_PARAM,
            ST         => UC (T.Sched_App).Scheduler,
            T          => T);
      end if;
   end New_Sched_Param_In_AppSched_Task;

   -------------------------------------------
   --  Trying_To_Change_Policy_To_Sched_App --
   -------------------------------------------
   procedure Trying_To_Change_Policy_To_Sched_App (T : in Task_Id) is
   begin
      pragma Assert (Task_OK (T),
                     "Trying_To_Change_Policy_To_Sched_App: invalid Task");

      --  Suspend changed task until appscheduler's response (unless change
      --  is being performed from the task itself)
      if T /= Running_Task and then T.Status = READY then
         Non_Running_Task_Gets_Blocked (T, WAITING_POLICY_CHANGE_TO_SCHED_APP);
      end if;
      --  Suspend task that changes the policy (if the NEW event is
      --  unmasked for the appscheduler) and send event to the appscheduler.
      APPSCHD.Send_Event_From_Task
        (Event_Code => APPSCHD_DATA.APPSCHED_NEW,
         ST         => UC (T.Sched_App).Scheduler,
         T          => T);
   end Trying_To_Change_Policy_To_Sched_App;


   -----------------
   -- Remove_Task --
   -----------------
   --
   --  Removes the task from the ready queue and also removes its asociated
   --  events from the timed events queue.
   procedure Remove_Task (T : in Task_Id) is
      Event_Ac : Kernel.Timed_Event_Ac;
      I : TEQ.Index;
   begin
      pragma Assert (Kernel.Task_Terminated (T), "Remove_Task: invalid Task");

      if T.Status = READY then
         ML.Dequeue (T, Ready_Queue);
      elsif T.Status = TIME_SUSPENDED or T.Status = TIMED_WAITING_CV then
         TE_T.Remove_Standard_Event (T.Suspension_Event);
      end if;
      case T.Sched_Policy is
         when SCHED_FIFO | SCHED_RR | SCHED_OTHER =>
            if T.AppScheduler /= null then
               --  Set to null the field 'App_Scheduler' of its scheduled task
               declare
                  TC : APPSCHD_DATA.Task_Container_Ac;
                  use APPSCHD_DATA.Task_Containers_Lists;
               begin
                  TC := Head (UC (T.AppScheduler).Tasks_Q);
                  while TC /= null loop
                     UC (TC.T.Sched_App).Scheduler := null;
                     TC := Next (TC);
                  end loop;
               end;
            end if;
         when SCHED_SPORADIC =>
            --  Look for the replenishment events of 'T' and remove them
            I := 1;
            loop
               Event_Ac := TEQ.Timed_Event_In_Pos (I);
               exit when Event_Ac = null;
               if ((Event_Ac.all
                    in Replenishment_TE.Replenishment_Timed_Event'Class)
                   and then
                   Replenishment_TE.Replenishment_Timed_Event_Ac
                   (Event_Ac).SS_Task = T) then
                  --  Remove the event. 'I' isn't incremented because when
                  --  an element is removed the rest of elements in the
                  --  'TEQ' are shifted to occupy the empty position.
                  TEQ.Remove (Event_Ac);
               else
                  exit when I = TEQ.Index'Last;
                  I := I + 1;
               end if;
            end loop;
         when SCHED_APP =>
            APPSCHD.Send_Event_From_Task
              (Event_Code => APPSCHD_DATA.APPSCHED_TERMINATE,
               ST         => UC (T.Sched_App).Scheduler,
               T          => T);
      end case;
   end Remove_Task;

   -----------------------------
   -- Running_Task_Yields_CPU --
   -----------------------------
   procedure Running_Task_Yields_CPU is
      use Option_Sets;
   begin
      if (Running_Task.Sched_Policy /= SCHED_APP or else
          APPSCHD_DATA.To_Option_Set (APPSCHD_DATA.APPSCHED_YIELD) <=
          UC (UC (Running_Task.Sched_App).Scheduler.AppScheduler).Events_Mask)
      then
         --  Non-SCHED_APP tasks or SCHED_APP tasks with the
         --  APPSCHED_YIELD event masked have to be reenqueued.
         if Running_Task.Reenqueuement_Deferred then
            Reenqueue_Task_Ending_Reenqueuement_Deferred (Running_Task);
         else
            ML.Reenqueue_Head_In_Order (Running_Task, Ready_Queue);
         end if;
      end if;
      case Running_Task.Sched_Policy is
         when SCHED_FIFO | SCHED_SPORADIC | SCHED_OTHER =>
            null;
         when SCHED_RR =>
            Dequeue (Running_Task.Scheduler_Timed_Event'Access,
                     Running_Task.CPU_Time_TEs_Q);
            Running_Task.Scheduler_Timed_Event.CPU_Time :=
              Running_Task.Scheduler_Timed_Event.CPU_Time
              + Running_Task.Capacity;
            Running_Task.Scheduler_Timed_Event.T :=
              Running_Task.Scheduler_Timed_Event.T + Running_Task.Capacity;
            Enqueue_In_Order (Running_Task.Scheduler_Timed_Event'Access,
                              Running_Task.CPU_Time_TEs_Q);
         when SCHED_APP =>
            APPSCHD.Send_Event_From_Task
              (Event_Code => APPSCHD_DATA.APPSCHED_YIELD,
               ST         => UC (Running_Task.Sched_App).Scheduler,
               T          => Running_Task);
      end case;
   end Running_Task_Yields_CPU;

   ----------------------------------------------------------
   --  Global variables used by 'Timer_Interrupt_Handler'  --
   --  and 'Do_Scheduling'                                 --
   ----------------------------------------------------------
   CS_Time : HWI.HWTime; -- Context switch time
   After_Timed_Event_Expiration : Boolean := False;

   -----------------------------
   -- Timer_Interrupt_Handler --
   -----------------------------
   procedure Timer_Interrupt_Handler
     (State : in Hardware_Interface.Trap_State_Ac);
   procedure Timer_Interrupt_Handler
     (State : in Hardware_Interface.Trap_State_Ac) is
      Tmp_Event : Timed_Event_Ac;
      Event_ST  : TSTE.Task_Suspension_TE_Ac;
      Timer_Event : Kernel.Timers.Timer_Id;
      Repl_Event : Kernel.Replenishment_TE.Replenishment_Timed_Event_Ac;
      type NTE_Ac is access all APPSCHD_DATA.Notification_Timed_Event;
   begin

      --  It is an error being here with interrupts enabled
      pragma Assert (not HWI.Are_Interrupts_Enabled);

      CS_Time := TE_T.Get_Timer_Activation_Time;

      Tmp_Event := TE_T.Extract_Expired_Event (Running_Task);

      if Tmp_Event = null then
         --  Periodic timer interrupt. Just reprogram the timer again.
         TE_T.Reprogram_Timer_After_Expiration (Running_Task);
         return;

      else
         loop
            After_Timed_Event_Expiration := True;


            if Tmp_Event.all in TSTE.Task_Suspension_Timed_Event'Class then

               ------------
               --  Task suspension timed event
               ----
               Event_ST  := TSTE.Task_Suspension_TE_Ac (Tmp_Event);

               pragma Assert (Task_OK (Event_ST.Suspended_Task));
               --  pragma Debug -- {CDBG}
               --    (CDbg.Timer_Interrupt_Handler_TSTE -- {CDBG}
               --     (Event_ST, CS_Time)); -- {CDBG}

               case Event_ST.Reason is
                  when TSTE.CV_TIMEDWAIT =>
                     Kernel.Condition_Variables.Internals.Timeout_Reached
                       (Event_ST.Suspended_Task, Event_ST.CV_Where_Blocked);

                  when TSTE.MUTEX_TIMEDWAIT =>
                     Kernel.Mutexes.Internals.Timeout_Reached
                       (Event_ST.Suspended_Task, Event_ST.Mutex_Where_Blocked);

                  when TSTE.SEM_TIMEDWAIT =>
                     Kernel.Semaphores.Internals.Timeout_Reached
                       (Event_ST.Suspended_Task, Event_ST.Sem_Where_Blocked);

                  when TSTE.SUSPENSION =>
                     Task_Gets_Ready (Event_ST.Suspended_Task);

                  when TSTE.APPSCHED_EVENT_TIMEDWAIT =>
                     APPSCHD.Send_Timeout_Event (Event_ST.Suspended_Task);

                  when TSTE.HWINTR_TIMEDWAIT =>
                     Kernel.Hardware_Interrupts.Timeout_Reached
                       (Event_ST.Suspended_Task);
               end case;

            elsif Tmp_Event = Running_Task.Scheduler_Timed_Event'Access then

               ------------
               --  Scheduler timed event (RR and SS)
               ----
               pragma Assert (Task_OK (Running_Task) and then
                              (Running_Task.Sched_Policy = SCHED_RR or else
                               Running_Task.Sched_Policy = SCHED_SPORADIC));
               --  pragma Debug -- {CDBG}
               --    (CDbg.Timer_Interrupt_Handler_STE -- {CDBG}
               --     (Running_Task, Tmp_Event,CS_Time)); -- {CDBG}
               case Running_Task.Sched_Policy is
                  when SCHED_RR =>
                     ------------
                     --  End of RR slice
                     ----
                     if not Running_Task.Reenqueuement_Deferred then
                        ML.Reenqueue_Head_In_Order (Running_Task, Ready_Queue);
                     end if;
                     --  If the 'Reenqueuement_Deferred' flag is set it isn't
                     --  necessary to do anything because it's sure there isn't
                     --  another task in the 'Running_Task' active priority
                     --  queue, otherwise the reenqueuement deferred would has
                     --  already been ended.
                     Running_Task.Scheduler_Timed_Event.CPU_Time :=
                       Running_Task.Scheduler_Timed_Event.CPU_Time
                       + Running_Task.Capacity;
                     Running_Task.Scheduler_Timed_Event.T :=
                       Running_Task.Scheduler_Timed_Event.T
                       + Running_Task.Capacity;
                     Enqueue_In_Order
                       (Running_Task.Scheduler_Timed_Event'Access,
                        Running_Task.CPU_Time_TEs_Q);

                  when SCHED_SPORADIC =>
                     ------------
                     --  SS Task has reached its execution time limit
                     ----

                     --  A replenishment operation is scheduled.
                     Schedule_SS_Replenishment_Operation
                       (Now => CS_Time, Limit_Reached => True);
                     --  The priority of the task is lowered.
                     TOI.Set_Base_Prio (Running_Task,
                                        Running_Task.Sched_SS_Low_Priority);
                     Running_Task.Capacity_Exhausted := True;

                  when others =>
                     --  scheduler event in not RR nor SS task!!
                     Kernel_Console.Error ("#70#");
               end case;

            elsif Tmp_Event.all in Kernel.Timers.Timer_Timed_Event'Class then

               --  pragma Debug -- {CDBG}
               --    (CDbg.Timer_Interrupt_Handler_TE -- {CDBG}
               --     (Kernel.Timers.Timer_Id (Tmp_Event), CS_Time)); -- {CDBG}

               ------------
               --  Timer timed event
               ----
               Timer_Event := Kernel.Timers.Timer_Id (Tmp_Event);
               pragma Assert (Kernel.Timers.Internals.Clock_Id_OK
                                (Timer_Event.Base_Clock));
               Kernel.Timers.Internals.Timer_Expired (Timer_Event, CS_Time);

            elsif Tmp_Event.all
              in Kernel.Replenishment_TE.Replenishment_Timed_Event'Class then

               ------------
               --  Sporadic Server Replenishment timed event
               ----
               Repl_Event :=
                 Replenishment_TE.Replenishment_Timed_Event_Ac (Tmp_Event);

               pragma Assert
                 (Task_OK (Repl_Event.SS_Task) and then
                  Repl_Event.SS_Task.Sched_Policy = SCHED_SPORADIC);
               --  pragma Debug -- {CDBG}
               --    (CDbg.Timer_Interrupt_Handler_RTE -- {CDBG}
               --     (Repl_Event, CS_Time)); -- {CDBG}

               --  Update the 'Number_Of_Pending_Replenishments' and the
               --  'Replenish_Amount_Pending'
               Repl_Event.SS_Task.Number_Of_Pending_Replenishments :=
                 Repl_Event.SS_Task.Number_Of_Pending_Replenishments - 1;
               if Repl_Event.SS_Task.Number_Of_Pending_Replenishments = 0 then
                  Repl_Event.SS_Task.Replenish_Amount_Pending := 0;
               else
                  Repl_Event.SS_Task.Replenish_Amount_Pending :=
                    Repl_Event.SS_Task.Replenish_Amount_Pending
                    - Repl_Event.Replenish_Amount;
               end if;

               --  Sporadic Server Task at high or low priority ?
               if not Repl_Event.SS_Task.Capacity_Exhausted then
                  --  Task at high priority (available capacity): the
                  --  scheduler timed event is going to be updated to expire
                  --  'Replenish_Amount' latter than currently. It must be
                  --  dequeued before doing this change.
                  CPU_Time_Timed_Events_Lists.Dequeue
                    (Repl_Event.SS_Task.Scheduler_Timed_Event'Access,
                     Repl_Event.SS_Task.CPU_Time_TEs_Q);
               else
                  --  Task at low priority (capacity exhausted): the
                  --  scheduler timed event is out of the queue with an
                  --  invalid value (it is going to be initialized with the
                  --  current values). Besides, the priority of the task is
                  --  raised and the activation time is stored.
                  if Repl_Event.SS_Task = Running_Task then
                     Running_Task.Scheduler_Timed_Event.CPU_Time :=
                       Running_Task.Used_CPU_Time +
                       (CS_Time - Running_Task.Running_Starting_Time);
                  else
                     Repl_Event.SS_Task.Scheduler_Timed_Event.CPU_Time :=
                       Repl_Event.SS_Task.Used_CPU_Time;
                  end if;
                  Repl_Event.SS_Task.Scheduler_Timed_Event.T := CS_Time;

                  TOI.Set_Base_Prio
                    (Repl_Event.SS_Task,
                     Repl_Event.SS_Task.Sched_SS_High_Priority);
                  Repl_Event.SS_Task.Capacity_Exhausted := False;
                  Repl_Event.SS_Task.Activation_Time := Repl_Event.T;

                  --  recarga
               end if;

               --  Add 'Replenish_Amount' to the scheduler timed event and
               --  enqueue it.
               Repl_Event.SS_Task.Scheduler_Timed_Event.CPU_Time :=
                 Repl_Event.SS_Task.Scheduler_Timed_Event.CPU_Time
                 + Repl_Event.Replenish_Amount;
               Repl_Event.SS_Task.Scheduler_Timed_Event.T :=
                 Repl_Event.SS_Task.Scheduler_Timed_Event.T
                 + Repl_Event.Replenish_Amount;

               CPU_Time_Timed_Events_Lists_Order.Enqueue_In_Order
                 (Repl_Event.SS_Task.Scheduler_Timed_Event'Access,
                  Repl_Event.SS_Task.CPU_Time_TEs_Q);

               Replenishment_TE.Release (Repl_Event);

            elsif Tmp_Event.all
              in APPSCHD_DATA.Notification_Timed_Event'Class then

               ------------
               --  SCHED_APP task notification event
               ----
               APPSCHD.Send_Notification_Or_Ready
                 (Event_Code => APPSCHD_DATA.APPSCHED_TASK_NOTIFICATION,
                  ST         =>
                    UC (NTE_Ac (Tmp_Event).Task_To_Notify.Sched_App).Scheduler,
                  T          => NTE_Ac (Tmp_Event).Task_To_Notify);
               NTE_Ac (Tmp_Event).Used := False;

            else
               --  Timed Event of a non-expected class !!
               Kernel_Console.Error ("#71#");

            end if;

            Tmp_Event := TE_T.Extract_Expired_Event (Running_Task);
            exit when Tmp_Event = null;

         end loop;
      end if;

      --  The "software interrupt handler" ('Do_Scheduling') is put pending.
      --  It will be called when all outstanding hardware interrupts have
      --  been processed.
      Interrupt_Nesting_Count :=
        Interrupt_Nesting_Count and Invoke_Do_Scheduling_Mask;

   end Timer_Interrupt_Handler;

   ---------------------------------
   -- Running_Task_Gets_Suspended --
   ---------------------------------
   procedure Running_Task_Gets_Suspended (With_Status : in Task_Status;
                                          At_Time : in HWI.HWTime) is
   begin
      Running_Task_Gets_Blocked (With_Status);
      --  Enqueue Timed event
      TE_T.Enqueue_Standard_Event_And_Update_Timer
        (Running_Task.Suspension_Event, At_Time);
   end Running_Task_Gets_Suspended;

   ------------------------
   -- Remove_Timed_Event --
   ------------------------
   procedure Remove_Timed_Event (E : in Timed_Event_Ac)
     renames TE_T.Remove_Standard_Event;

   -------------------
   -- Do_Scheduling --
   -------------------
   Heir_Task, Old_Task : Task_Id;   --  Local to 'Do_Scheduling'.
   CPU_E : CPU_Time_Timed_Event_Ac; --  Local to 'Do_Scheduling'.
   procedure Do_Scheduling is
      --  Without local variables the execution time of this procedure is
      --  faster (no allocation of variables in the stack).
      use APPSCHD_DATA;
   begin
      --  Should be called with interrupts disabled
      pragma Assert (not HWI.Are_Interrupts_Enabled,
                     "Do_Scheduling: Interrupts Enabled");
      if Interrupt_Nesting_Count /= Outside_Of_Interrupts then
         --  The context switch mustn't be performed inside an interrupt
         --  handler. We just clear the high bit in
         --  'Interrupt_Nesting_Count' to put a "software interrupt"
         --  pending. 'Do_Scheduling' (the software interrupt handler) will
         --  be called when all outstanding hardware interrupts have been
         --  processed.
         Interrupt_Nesting_Count :=
           Interrupt_Nesting_Count and Invoke_Do_Scheduling_Mask;
         return;
      end if;

      ------------
      --  Reenqueuement Deferred ?
      ----
      if Running_Task.Reenqueuement_Deferred then
         pragma Assert (not HWI.Are_Interrupts_Enabled and
                          Running_Task.Sched_Policy /= SCHED_APP);
         --  pragma Debug (CDbg.Do_Scheduling1 (Running_Task)); -- {CDBG}

         Reenqueue_Task_Ending_Reenqueuement_Deferred (Running_Task);
      end if;

      Heir_Task := ML.Head (Ready_Queue);

      ------------
      --  Is the scheduler of the heir task ready ? (SCHED_APP)
      ----
      if (Heir_Task.Sched_Policy = SCHED_APP and then
          UC (Heir_Task.Sched_App).Scheduler /= null and then
          UC (Heir_Task.Sched_App).Scheduler.Status = READY) then
         --  An application scheduled task should never run before its
         --  scheduler in the case it is READY. When this happens the
         --  scheduler is moved from its current position in the ready
         --  queue to the head and it becomes the heir task.
         Heir_Task := UC (Heir_Task.Sched_App).Scheduler;
         ML.Dequeue (Heir_Task, Ready_Queue);
         ML.Enqueue_First_In_Priority (Heir_Task, Ready_Queue);
      end if;

      ------------
      --  Has the most prioritary ready task changed ?
      ----
      if Running_Task /= Heir_Task then
         Old_Task := Running_Task;
         Running_Task := Heir_Task;
         pragma Debug (Debug.Assert_Tasks_Are_OK (Heir => Heir_Task,
                                                  Old  => Old_Task));

         if not After_Timed_Event_Expiration then
            CS_Time := HWI.Get_HWTime;
         end if;

         --  pragma Debug -- {CDBG}
         --    (CDbg.Do_Scheduling2 -- {CDBG}
         --       (Old_Task, Heir_Task, Ready_Queue, CS_Time)); -- {CDBG}

         ------------
         --  CPU times measurements
         ----
         if (Configuration_Parameters.Use_CPU_Time_Clocks_And_Timers    or else
             Configuration_Parameters.Use_Round_Robin_Scheduling_Policy or else
             Configuration_Parameters.Use_Sporadic_Server_Scheduling_Policy)
         then
            Heir_Task.Running_Starting_Time := CS_Time;
            Old_Task.Used_CPU_Time := Old_Task.Used_CPU_Time +
              (CS_Time - Old_Task.Running_Starting_Time);

            --  Gets absolute time of heir task's CPU Time Timed Events
            CPU_E := Head (Heir_Task.CPU_Time_TEs_Q);
            while CPU_E /= null loop
               CPU_E.T := CS_Time + (CPU_E.CPU_Time - Heir_Task.Used_CPU_Time);
               CPU_E := Next (CPU_E);
            end loop;
         end if;

         --  Timer reprogramming
         if After_Timed_Event_Expiration then
            After_Timed_Event_Expiration := False;
            TE_T.Reprogram_Timer_After_Expiration (Running_Task);
         elsif (Configuration_Parameters.Use_CPU_Time_Clocks_And_Timers
                or else
                Configuration_Parameters.Use_Round_Robin_Scheduling_Policy
                or else
                Configuration_Parameters.Use_Sporadic_Server_Scheduling_Policy)
         then
            --  The the new running task's most urgent CPU time event could
            --  be the most urgent event.
            TE_T.Update_Timer_Due_To_New_Running_Task (Running_Task, CS_Time);
         end if;

         ------------
         --  The context switch is actually carried out here
         ----
         HWI.Context_Switch
           (Old_Task => Old_Task.Stack_Ptr'Address,
            New_Task => Heir_Task.Stack_Ptr'Address);
      elsif After_Timed_Event_Expiration then

         After_Timed_Event_Expiration := False;
         TE_T.Reprogram_Timer_After_Expiration (Running_Task);

      end if;
   end Do_Scheduling;
   pragma Inline (Do_Scheduling);

   --------------------
   -- Idle_Task_Body --
   --------------------
   function Idle_Task_Body (Arg : System.Address) return System.Address;
   function Idle_Task_Body (Arg : System.Address) return System.Address is
   begin
      while True loop
         --  pragma Assert (HWI.Are_Interrupts_Enabled);
         null;
      end loop;
      return System.Null_Address;
   end Idle_Task_Body;

begin
   ------------
   --  Initialize the Ready queue
   ----
   ML.Initialize (Ready_Queue);

   ------------
   --  Initialize and Register the main and idle task
   ----
   --
   --  Their respectives TCB are initialized in
   --  'Kernel.Tasks_Operations.Internals'
   Running_Task :=
     TO.Initialize_TCBs.Initialize_Main_Task (Kernel.Task_Priority'Last / 2);
   Register_Main_Task (Running_Task);

   Register_Idle_Task
     (TO.Initialize_TCBs.Initialize_Idle_Task (Idle_Task_Body'Access,
                                               TOI.Task_Wrapper'Address));

   ------------
   --  Initialize user interrupt handlers
   ----
   if Configuration_Parameters.Use_HW_Interrupts_Control'First then
      Kernel.Hardware_Interrupts.Initialize_User_Interrupt_Handlers;
   end if;

   ------------
   --  Install the Timer interrupt handler
   ----
   HWI.Install_HW_Interrupt_Handler (HWI.TIMER_INTERRUPT,
                                     Timer_Interrupt_Handler'Access);
   Kernel.Hardware_Interrupts.Reserve_Interrupt (HWI.TIMER_INTERRUPT);
   HWI.Interrupt_Controller_Hardware_Enable_Interrupt (HWI.TIMER_INTERRUPT);

   --  After elaborating this package will be elaborated
   --  'Kernel.Devices_Table.Init'. It is in this last package where
   --  the devices are initialized and, after that, interrupts are
   --  enabled.

end Kernel.Scheduler;
