------------------------------------------------------------------------------
--  ------------------         M a R T E   O S         -------------------  --
------------------------------------------------------------------------------
--                                                             V1.57  Dic 2005
--
--               'K e r n e l . T a s k s _ O p e r a t i o n s'
--
--                                 Body
--
--
--  File 'k-tasks_operation.adb'                                      By MAR.
--
--
--  Operations with Tasks.
--
--  ----------------------------------------------------------------------
--   Copyright (C) 2004   Universidad de Cantabria, SPAIN
--
--   MaRTE OS web page: http://marte.unican.es
--   Contact Addresses: Mario Aldea Rivas          aldeam@unican.es
--                      Michael González Harbour      mgh@unican.es
--
--  MaRTE OS  is free software; you can  redistribute it and/or  modify it
--  under the terms of the GNU General Public License  as published by the
--  Free Software Foundation;  either  version 2, or (at  your option) any
--  later version.
--
--  MaRTE OS  is distributed  in the  hope  that  it will be   useful, but
--  WITHOUT  ANY  WARRANTY;     without  even the   implied   warranty  of
--  MERCHANTABILITY  or  FITNESS FOR A  PARTICULAR PURPOSE.    See the GNU
--  General Public License for more details.
--
--  You should have received  a  copy of  the  GNU General Public  License
--  distributed with MaRTE  OS;  see file COPYING.   If not,  write to the
--  Free Software  Foundation,  59 Temple Place  -  Suite 330,  Boston, MA
--  02111-1307, USA.
--
--  As a  special exception, if you  link this  unit  with other  files to
--  produce an   executable,   this unit  does  not  by  itself cause  the
--  resulting executable to be covered by the  GNU General Public License.
--  This exception does  not however invalidate  any other reasons why the
--  executable file might be covered by the GNU Public License.
--
------------------------------------------------------------------------------
with Hardware_Interface; use Hardware_Interface;
with Configuration_Parameters;
with Kernel.Scheduler;
with Kernel.Application_Scheduling_Data;
with Kernel.Mutexes;

with Kernel.Tasks_Operations.Internals;

--  Debug
with Kernel_Console;

package body Kernel.Tasks_Operations is

   package HWI          renames Hardware_Interface;
   package SCHD         renames Kernel.Scheduler;
   package CP           renames Configuration_Parameters;
   package APPSCHD_DATA renames Kernel.Application_Scheduling_Data;

   use Kernel.Scheduler, Kernel.Tasks_Lists;

   function UC (Base : Kernel.Sched_App_Data_Base_Ac)
                return APPSCHD_DATA.Sched_App_Data_Ac
     renames APPSCHD_DATA.UC_To_Sched_App_Data_Ac;
   pragma Inline (UC);

   -------------------
   -- Set_Base_Prio --
   -------------------
   procedure Set_Base_Prio (T             : in Task_Id;
                            New_Base_Prio : in Task_Priority);
   procedure Set_Base_Prio (T             : in Task_Id;
                            New_Base_Prio : in Task_Priority) is
      Flags : Integer;
   begin
      Kernel.Enter_Critic_Section (Flags);
      if New_Base_Prio /= T.Active_Prio then
         T.Base_Prio := New_Base_Prio;
         if Mutexes_Lists.Is_Empty (T.Mutexes_Owned) then
            if New_Base_Prio > T.Active_Prio then
               Internals.Raise_Active_Prio (T, New_Base_Prio);
            else
               Internals.Reduce_Active_Prio (T, New_Base_Prio,
                                             Head_Of_New_Prio_Q => False);
            end if;
         else -- not Mutexes_Lists.Is_Empty (Mutexes_Owned)
            if New_Base_Prio > T.Active_Prio then
               Internals.Raise_Active_Prio (T, New_Base_Prio);
            end if;
         end if;
      end if;
      Kernel.Leave_Critic_Section (Flags);
   end Set_Base_Prio;


   ----------------------------------------------------------------------------
   -- Get and Set Scheduling Parameters and Policy ----------------------------
   ----------------------------------------------------------------------------

   ------------------------------------
   -- Error_In_Scheduling_Parameters --
   ------------------------------------
   --
   --  Check the 'Policy' and 'Parameters'. Called from
   --  'Pthread_Create', 'TO_ATTR.Pthread_Attr_Setschedparam' and
   --  'Pthread_Setschedparam'.
   function Error_In_Scheduling_Parameters (Param  : in Sched_Param;
                                            Policy : in Scheduling_Policies)
                                            return Kernel.Error_Code is
   begin
      --  Functionality supported
      if ((not Configuration_Parameters.Use_Round_Robin_Scheduling_Policy
           and then Policy = SCHED_RR) or
          (not Configuration_Parameters.Use_Sporadic_Server_Scheduling_Policy
           and then Policy = SCHED_SPORADIC) or
          (not Configuration_Parameters.Use_Application_Defined_Scheduling
           and then Policy = SCHED_APP))then
         return OPERATION_NOT_SUPPORTED;
      end if;
      --  Priority range
      if (Param.Sched_Priority < Task_Priority'First or
          Param.Sched_Priority > Task_Priority'Last) then
         return INVALID_ARGUMENT;
      end if;
      --  Policy range
      if (Policy < Scheduling_Policies'First or
          Policy > Scheduling_Policies'Last) then
         return INVALID_ARGUMENT;
      end if;
      --  Policy dependent errors
      case Policy is
         when SCHED_APP => --  SCHED_APP related errors
            null;

         when SCHED_SPORADIC => --  SCHED_SPORADIC related errors
            if (Type_Timespec.Timespec_To_HWTime (Param.Sched_SS_Repl_Period) <
                Type_Timespec.Timespec_To_HWTime (Param.Sched_SS_Init_Budget))
            then
               --  Invalid repl. period or budget
               return INVALID_ARGUMENT;
            end if;
            if (Param.Sched_SS_Max_Repl > CP.Sporadic_Server_Replenishments_Mx
                or Param.Sched_SS_Max_Repl <= 0) then
               --  Invalid Max. number of simultaneous replenishment operations
               return INVALID_ARGUMENT;
            end if;

         when SCHED_FIFO | SCHED_RR | SCHED_OTHER  =>
            null;
      end case;
      return NO_ERROR;
   end Error_In_Scheduling_Parameters;

   ---------------------------
   -- Pthread_Getschedparam --
   ---------------------------
   function Pthread_Getschedparam (T      : in Task_Id;
                                   Policy : access Scheduling_Policies;
                                   Param  : access Sched_Param)
                                   return Int is
   begin
      Policy.all := T.Sched_Policy;

      Param.all := Sched_Param'
        (Sched_Priority      => T.Base_Prio,
         Sched_SS_Low_Priority => T.Sched_SS_Low_Priority,
         Sched_SS_Repl_Period  =>
           Type_Timespec.HWTime_To_Timespec (T.Sched_SS_Repl_Period),
         Sched_SS_Init_Budget  =>
           Type_Timespec.HWTime_To_Timespec (T.Capacity),
         Sched_SS_Max_Repl     => T.Sched_SS_Max_Repl);

      return 0;
   end Pthread_Getschedparam;

   ---------------------------
   -- Pthread_Setschedparam --
   ---------------------------
   function Pthread_Setschedparam (T      : in     Task_Id;
                                   Policy : in     Scheduling_Policies;
                                   Param  : access Sched_Param)
                                   return Int is
      Flags : Integer;
      Original_Base_Priority : Kernel.Task_Priority;
      Changing_From_Standard_To_Sched_App : Boolean := False;
      Error : Kernel.Error_Code;
   begin
      Kernel.Enter_Critic_Section (Flags);
      if not Task_OK (T) then
         Kernel.Leave_Critic_Section (Flags);
         return INVALID_ARGUMENT;
      end if;
      if (Configuration_Parameters.Use_Application_Defined_Scheduling and then
          Policy = SCHED_APP) then
         if T.Sched_Policy /= SCHED_APP and T.Status /= READY then
            --  Only for READY tasks is allowed changing to SCHED_APP policy
            Kernel.Leave_Critic_Section (Flags);
            return OPERATION_NOT_SUPPORTED;
         end if;
         if (not Task_OK (UC (T.Sched_App).Scheduler) or else
             UC (T.Sched_App).Scheduler.AppScheduler = null) then
            --  Invalid "appscheduler" attribute
            Kernel.Leave_Critic_Section (Flags);
            return INVALID_ARGUMENT;
         end if;
         if Param.Sched_Priority > UC (T.Sched_App).Scheduler.Base_Prio then
            --  Priority higher than its scheduler task
            Kernel.Leave_Critic_Section (Flags);
            return INVALID_ARGUMENT;
         end if;
      end if; -- Configuration_Parameters.Use_Application_Defined_Scheduling
      if (T.Sched_Policy = SCHED_SPORADIC and then
          (Policy = SCHED_SPORADIC and
           T.Number_Of_Pending_Replenishments /= 0)) then
         --  Only allowed changing SCHED_SPORADIC task's parameters when
         --  there are no pending replenishment operations
         Kernel.Leave_Critic_Section (Flags);
         return OPERATION_NOT_SUPPORTED;
      end if;
      Error := Error_In_Scheduling_Parameters (Param.all, Policy);
      if Error /= NO_ERROR then
         return Error;
      end if;

      --  Set task new parameters
      case Policy is
         when SCHED_FIFO | SCHED_OTHER =>
            null;

         when SCHED_RR =>
            if Configuration_Parameters.Use_Round_Robin_Scheduling_Policy then
               T.Capacity := HWI.Duration_To_HWTime
                 (CP.Round_Robin_Interval_Time_Period);
            end if;

         when SCHED_APP =>
            if (Configuration_Parameters.Use_Application_Defined_Scheduling
                and then T.Sched_Policy /= SCHED_APP) then
               --  Change from standard policy to SCHED_APP: before
               --  actually changing the policy the appscheduler must
               --  accept the new task.
               Changing_From_Standard_To_Sched_App := True;

               if not APPSCHD_DATA.Any_Task_Container_Free_In_Pool then
                  Kernel.Leave_Critic_Section (Flags);
                  return RESOURCE_TEMPORARILY_UNAVAILABLE;
               end if;

               --  Annotate task waiting appscheduler response
               T.Task_Waiting_Appscheduler_Response := TCB_Ac (Self);
               --  Set new base prio
               Original_Base_Priority := T.Base_Prio;
               T.Base_Prio := Param.Sched_Priority;

               --  Suspend task(s) and send NEW event to the scheduler
               SCHD.Trying_To_Change_Policy_To_Sched_App (T);
               SCHD.Do_Scheduling;
               --  When here is because the appscheduler has responded

               pragma Assert
                 (SCHD.Self.Internal_Error_Code = NO_ERROR or
                    SCHD.Self.Internal_Error_Code = APPSCHED_EVENT_MASKED or
                    SCHD.Self.Internal_Error_Code = APPSCHED_REJECTED);
               if SCHD.Self.Internal_Error_Code /= NO_ERROR then
                  --  Task cannot change its policy: the appscheduler has
                  --  the NEW_TASK event masked or task has been rejected
                  --  by its appscheduler
                  declare
                     Error : Kernel.Error_Code := Self.Internal_Error_Code;
                  begin
                     SCHD.Self.Internal_Error_Code := NO_ERROR;  --  reset flag
                     T.Base_Prio := Original_Base_Priority;
                     T.Task_Waiting_Appscheduler_Response := null;
                     if (T /= SCHD.Self and then
                         T.Status = WAITING_POLICY_CHANGE_TO_SCHED_APP) then
                        SCHD.Task_Gets_Ready (T);
                     end if;
                     Kernel.Leave_Critic_Section (Flags);
                     return Error;
                  end;
               else
                  --  NO_ERROR: give correct values to 'Sched_App'
                  --  fields. Some of them do not change since they
                  --  must have been set before this function is
                  --  called.
                  APPSCHD_DATA.Sched_App_Data (T.Sched_App.all) :=
                    (Scheduler              => UC (T.Sched_App).Scheduler,
                     Param                  => UC (T.Sched_App).Param,
                     Param_Size             => UC (T.Sched_App).Param_Size,
                     AppMutex_Where_Waiting => Mutexes.Null_Mutex_Descriptor,
                     Scheduler_Reply        => null,
                     Reply_Size             => 0,
                     Owner                  => T,
                     Notification_Event     =>
                       APPSCHD_DATA.Null_Notification_TE);
               end if;
            end if;

         when SCHED_SPORADIC =>
            if CP.Use_Sporadic_Server_Scheduling_Policy then
               T.Sched_SS_High_Priority := Param.Sched_Priority;
               T.Sched_SS_Low_Priority  := Param.Sched_SS_Low_Priority;
               T.Sched_SS_Repl_Period   :=
                 Type_Timespec.Timespec_To_HWTime (Param.Sched_SS_Repl_Period);
               T.Capacity               :=
                 Type_Timespec.Timespec_To_HWTime (Param.Sched_SS_Init_Budget);
               T.Sched_SS_Max_Repl      := Param.Sched_SS_Max_Repl;
               T.Replenish_Amount_Pending         := 0;
               T.Number_Of_Pending_Replenishments := 0;
            end if;
      end case;

      if not Changing_From_Standard_To_Sched_App then
         --  (When a task is changed from any standard policy to SCHED_APP
         --  these operations are performed in
         --  'TO_APPSCHD.Execute_Actions')

         --  Set new base priority
         Internals.Set_Base_Prio (T, Param.Sched_Priority);

         if Policy = SCHED_APP and T.Sched_Policy = SCHED_APP then
            --  Send event CHANGE_SCHED_PARAM to appscheduler
            SCHD.New_Sched_Param_In_AppSched_Task (T);
         else
            --  Inform the scheduler a task has changed its policy
            SCHD.Task_Changes_Policy (T, Policy);
            T.Sched_Policy := Policy;
         end if;

         Scheduler.Do_Scheduling;
      end if;

      Kernel.Leave_Critic_Section (Flags);
      return 0;
   end Pthread_Setschedparam;

   --------------------------
   -- Pthread_Setschedprio --
   --------------------------
   function Pthread_Setschedprio (T    : in Task_Id;
                                  Prio : in Basic_Integer_Types.Int)
                                 return Int is
      Flags : Integer;
   begin
      if not CP.Use_POSIX_Thread_Execution_Scheduling'First then
         return OPERATION_NOT_IMPLEMENTED;
      end if;
      Kernel.Enter_Critic_Section (Flags);
      if not Task_OK (T) then
         Kernel.Leave_Critic_Section (Flags);
         return NO_SUCH_PROCESS;
      end if;
      if (Prio < Int (Task_Priority'First) or
          Prio > Int (Task_Priority'Last))    then
         Kernel.Leave_Critic_Section (Flags);
         return INVALID_ARGUMENT;
      end if;
      if (T.Sched_Policy = SCHED_SPORADIC and then
          T.Number_Of_Pending_Replenishments /= 0) then
         --  Only allowed changing SCHED_SPORADIC task's parameters when
         --  there are no pending replenishment operations
         Kernel.Leave_Critic_Section (Flags);
         return OPERATION_NOT_SUPPORTED;
      end if;

      --  Set new base priority
      Internals.Set_Base_Prio (T, Task_Priority (Prio),
                               Head_Of_New_Prio_Q => True);

      if T.Sched_Policy = SCHED_APP then
         --  Send event CHANGE_SCHED_PARAM to appscheduler
         SCHD.New_Sched_Param_In_AppSched_Task (T);
      end if;

      Scheduler.Do_Scheduling;

      Kernel.Leave_Critic_Section (Flags);
      return 0;
   end Pthread_Setschedprio;


   -------------------------------------------------------------------------
   --  Basic Operations  ---------------------------------------------------
   -------------------------------------------------------------------------

   --------------------
   -- Pthread_Create --
   --------------------
   function Pthread_Create (T         : access Task_Id;
                            Attr      : Pthread_Attr_T_Ac_Const;
                            Task_Body : Task_Body_Function;
                            Arg       : System.Address) return Int is
      Flags : Integer;
      Error : Kernel.Error_Code;
   begin
      Kernel.Enter_Critic_Section (Flags);
      --  Error handling
      if Attr /= null and then not Internals.Attr_OK (Attr) then
         Kernel.Leave_Critic_Section (Flags);
         return INVALID_ARGUMENT;
      end if;
      if (Configuration_Parameters.Use_Application_Defined_Scheduling and then
          Attr /= null and then
          Attr.Sched_Policy = SCHED_APP) then
         if (not Task_OK (Attr.App_Scheduler) or else
             Attr.App_Scheduler.AppScheduler = null) then
            --  Invalid "appscheduler" attribute
            Kernel.Leave_Critic_Section (Flags);
            return INVALID_ARGUMENT;
         end if;
         if Attr.Param.Sched_Priority > Attr.App_Scheduler.Base_Prio then
            --  Priority higher than its scheduler task
            Kernel.Leave_Critic_Section (Flags);
            return INVALID_ARGUMENT;
         end if;
      end if;

      if Attr = null then
         T.all := Internals.Create_Task (Task_Body, Arg, Attr_Deffault);
      else
         Error := Error_In_Scheduling_Parameters (Attr.Param,
                                                  Attr.Sched_Policy);
         if Error /= NO_ERROR then
            Kernel.Leave_Critic_Section (Flags);
            return Error;
         end if;
         if Attr.Inherit_Sched = PTHREAD_INHERIT_SCHED then
            --  Inherits father attributes
            T.all := Internals.Create_Task (Task_Body, Arg,
                                            Internals.Get_Self_Attributes);

         else
            T.all := Internals.Create_Task (Task_Body, Arg, Attr.all);
         end if;
      end if;

      --  Error handling
      if Kernel.Tasks_Lists."=" (T.all, null) then
         Kernel.Leave_Critic_Section (Flags);
         return RESOURCE_TEMPORARILY_UNAVAILABLE;
      end if;

      Scheduler.New_Task_Created (T.all);
      Scheduler.Do_Scheduling;

      if (Configuration_Parameters.Use_Application_Defined_Scheduling and then
          T.all.Sched_Policy = SCHED_APP) then
         --  Here continues the "father" task.

         pragma Assert
           (SCHD.Self.Internal_Error_Code   = NO_ERROR or
              SCHD.Self.Internal_Error_Code = APPSCHED_EVENT_MASKED or
              SCHD.Self.Internal_Error_Code = APPSCHED_REJECTED);
         if SCHD.Self.Internal_Error_Code /= NO_ERROR then
            --  The appscheduler has the NEW_TASK event masked or
            --  created task has been rejected by its appscheduler: task
            --  cannot be created and its TCB is Released
            declare
               Error : Kernel.Error_Code := SCHD.Self.Internal_Error_Code;
            begin
               SCHD.Self.Internal_Error_Code := NO_ERROR;  --  reset flag
               T.all.Magic := TERMINATED;
               Internals.Remove_Task (T.all);
               Kernel.Leave_Critic_Section (Flags);
               return Error;
            end;
         end if;
      end if; -- SCHED_APP

      Kernel.Leave_Critic_Section (Flags);
      return 0;
   end Pthread_Create;

   ------------------
   -- Pthread_Exit --
   ------------------
   procedure Pthread_Exit (Value_Ptr : in System.Address) is
      Flags : Integer;
   begin
      Kernel.Enter_Critic_Section (Flags);
      Internals.Terminate_Running_Task (Value_Ptr);
      Kernel.Leave_Critic_Section (Flags);
      --  This line should never be reached
      Kernel_Console.Error ("#61#", Fatal => True);
   end Pthread_Exit;

   ------------------
   -- Pthread_Join --
   ------------------
   function Pthread_Join (T : Task_Id; Value_Ptr : System_Address_Ac)
                          return Int is
      Flags : Integer;
   begin
      Kernel.Enter_Critic_Section (Flags);
      if not Task_OK (T, TERMINATED) then
         Kernel.Leave_Critic_Section (Flags);
         return NO_SUCH_PROCESS;
      end if;
      if T.Detach_State = PTHREAD_CREATE_DETACHED then
         Kernel.Leave_Critic_Section (Flags);
         return INVALID_ARGUMENT;
      end if;
      if T = Self or else Self.Task_Waiting_For_Termination = TCB_Ac (T) then
         Kernel.Leave_Critic_Section (Flags);
         return RESOURCE_DEADLOCK_AVOIDED;
      end if;

      if T.Magic /= TERMINATED then
         --  Wait for the target task termination
         T.Task_Waiting_For_Termination := TCB_Ac (Self);
         Self.Task_Whose_Termination_Is_Waited := TCB_Ac (T);
         SCHD.Running_Task_Gets_Blocked (WAITING_TASK_TERMINATION);
         SCHD.Do_Scheduling;

         --  Reactivation
         Self.Task_Whose_Termination_Is_Waited := null;
         --  Check the reason for the task reactivation
         pragma Assert (SCHD.Self.Internal_Error_Code = NO_ERROR or
                          SCHD.Self.Internal_Error_Code = INVALID_ARGUMENT);
         if SCHD.Self.Internal_Error_Code = INVALID_ARGUMENT then
            --  Some other task has call 'Pthread_Detach' for the target
            --  task
            SCHD.Self.Internal_Error_Code := NO_ERROR;  --  reset flag
            Kernel.Leave_Critic_Section (Flags);
            return INVALID_ARGUMENT;
         end if;
      end if;

      --  The target task is terminated.
      if Value_Ptr /= null then
         Value_Ptr.all := T.Ret_Val;
      end if;
      Internals.Remove_Task (T);

      Kernel.Leave_Critic_Section (Flags);
      return 0;
   end Pthread_Join;

   --------------------
   -- Pthread_Detach --
   --------------------
   function Pthread_Detach (T : Task_Id) return Int is
      Flags : Integer;
   begin
      Kernel.Enter_Critic_Section (Flags);
      if not Task_OK (T, TERMINATED) then
         Kernel.Leave_Critic_Section (Flags);
         return NO_SUCH_PROCESS;
      end if;
      if T.Detach_State = PTHREAD_CREATE_DETACHED then
         Kernel.Leave_Critic_Section (Flags);
         return INVALID_ARGUMENT;
      end if;

      T.Detach_State := PTHREAD_CREATE_DETACHED;

      --  Is there some task waiting for 'T' termination ?
      if T.Task_Waiting_For_Termination /= null then
         T.Task_Waiting_For_Termination.Task_Whose_Termination_Is_Waited :=
           null;
         --  Indicate error in the 'Internal_Notification' flag
         pragma Assert (T.Internal_Error_Code = NO_ERROR);
         pragma Assert
           (TCB_Ac_To_Task_Id
              (T.Task_Waiting_For_Termination).Internal_Error_Code = NO_ERROR);
         TCB_Ac_To_Task_Id
           (T.Task_Waiting_For_Termination).Internal_Error_Code :=
           INVALID_ARGUMENT;
         --  Put waiting task ready
         SCHD.Task_Gets_Ready
           (TCB_Ac_To_Task_Id (T.Task_Waiting_For_Termination));
         SCHD.Do_Scheduling;
      end if;

      Kernel.Leave_Critic_Section (Flags);
      return 0;
   end Pthread_Detach;

   -----------
   -- Yield --
   -----------
   function Sched_Yield return Int is
      Flags : Integer;
   begin
      Kernel.Enter_Critic_Section (Flags);
      --  Reenqueue occuping the last position in its priority
      Scheduler.Running_Task_Yields_CPU;

      --  Call Scheduler
      Scheduler.Do_Scheduling;

      Kernel.Leave_Critic_Section (Flags);
      return 0;
   end Sched_Yield;

end Kernel.Tasks_Operations;
