------------------------------------------------------------------------------
--  ------------------         M a R T E   O S         -------------------  --
------------------------------------------------------------------------------
--                                                             V1.57  Dic 2005
--
--             'K e r n e l . T i m e d _ E v e n t s _ Q u e u e'
--
--                                  Body
--
--
--  File 'k-timed_events_queue.ads'                                    By MAR.
--
--
--  Timed Events Queue Management. This package keeps a ordered binary
--  tree of timed events. This tree is used by 'Kernel.Timed_Events' to
--  store all the timed events but the CPU time events (including the
--  round-robin and sporadic server ones).
--
--  ----------------------------------------------------------------------
--   Copyright (C) 2004   Universidad de Cantabria, SPAIN
--
--   MaRTE OS web page: http://marte.unican.es
--   Contact Addresses: Mario Aldea Rivas          aldeam@unican.es
--                      Michael González Harbour      mgh@unican.es
--
--  MaRTE OS  is free software; you can  redistribute it and/or  modify it
--  under the terms of the GNU General Public License  as published by the
--  Free Software Foundation;  either  version 2, or (at  your option) any
--  later version.
--
--  MaRTE OS  is distributed  in the  hope  that  it will be   useful, but
--  WITHOUT  ANY  WARRANTY;     without  even the   implied   warranty  of
--  MERCHANTABILITY  or  FITNESS FOR A  PARTICULAR PURPOSE.    See the GNU
--  General Public License for more details.
--
--  You should have received  a  copy of  the  GNU General Public  License
--  distributed with MaRTE  OS;  see file COPYING.   If not,  write to the
--  Free Software  Foundation,  59 Temple Place  -  Suite 330,  Boston, MA
--  02111-1307, USA.
--
--  As a  special exception, if you  link this  unit  with other  files to
--  produce an   executable,   this unit  does  not  by  itself cause  the
--  resulting executable to be covered by the  GNU General Public License.
--  This exception does  not however invalidate  any other reasons why the
--  executable file might be covered by the GNU Public License.
--
------------------------------------------------------------------------------

with Hardware_Interface; use Hardware_Interface;
pragma Elaborate (Hardware_Interface);
with Configuration_Parameters;

--  Debug
with Basic_Console_IO;
with Debug_Messages;
with Kernel.Debug;

package body Kernel.Timed_Events_Queue is

   use Kernel.Timed_Events_Lists;

   procedure Show_TEQ_On_Console (Msg : in String);

   ------------------
   -- Events Queue --
   ------------------
   subtype Size_T is Integer
     range 0 .. Configuration_Parameters.Num_Timed_Events_Mx;

   Q    : array (Index) of Timed_Event_Ac;
   Size : Size_T;

   procedure UpHeap (N : in Index);
   procedure UpHeap (N : in Index) is
      P : Index;
      K : Index := N;
      Temp : Timed_Event_Ac;
   begin
      while K >= 2 loop
         P := K / 2;
         if Q (K).T < Q (P).T then
            Temp := Q (K);
            Q (K) := Q (P);
            Q (P) := Temp;
            K := P;
         else
            exit;
         end if;
      end loop;
   end UpHeap;
   pragma Inline (UpHeap);

   ----------------------
   -- Enqueue_In_Order --
   ----------------------
   procedure Enqueue_In_Order (E : in Timed_Event_Ac) is
   begin
      pragma Assert (Size < Index'Last,
                     "Enqueue_In_Order: queue maximum size reached");
      if Size < Index'Last then
         Size := Size + 1;
         Q (Size) := E;
         UpHeap (Size);
      end if;
      pragma Debug (Show_TEQ_On_Console ("Enq. Event EQ:"));
   end Enqueue_In_Order;

   procedure DownHeap (K : in Index; N : in Size_T);
   procedure DownHeap (K : in Index; N : in Size_T) is
      J : Integer := 2 * K;
      Half : Index := K;
      Temp : Timed_Event_Ac := Q (K);
   begin
      while J <= N loop
         if (J < N) and then (Q (J + 1).T < Q (J).T) then
            J := J + 1;
         end if;
         if not (Q (J).T < Temp.T) then
            Q (Half) := Temp;
            return;
         end if;
         Q (Half) := Q (J);
         Half := J;
         J := 2 * J;
      end loop;
      Q (Half) := Temp;
   end DownHeap;
   pragma Inline (DownHeap);


   -----------------
   -- Remove_Head --
   -----------------
   function Remove_Head return Timed_Event_Ac is
      E : Timed_Event_Ac;
   begin
      if Size = 0 then
         return null;
      end if;
      E := Q (1);
      Q (1) := Q (Size);
      Size := Size - 1;
      DownHeap (1, Size);
      pragma Debug (Show_TEQ_On_Console ("RM head EQ:"));
      return E;
   end Remove_Head;

   ------------
   -- Remove --
   ------------
   procedure Remove (E : in Timed_Event_Ac) is
      I : Integer := 0;
   begin
      while I < Size loop
         I := I + 1;
         if E = Q (I) then
            Q (I) := Q (Size);
            Size := Size - 1;
            DownHeap (I, Size);
            pragma Debug (Show_TEQ_On_Console ("RM EQ:"));
            return;
         end if;
      end loop;
      pragma Assert (False, "Remove: event is not in The queue");
   end Remove;


   ----------
   -- Head --
   ----------
   function Head return Timed_Event_Ac is

   begin
      if Size = 0 then
         return null;
      else
         return Q (1);
      end if;
   end Head;

   --------------
   -- Is_Empty --
   --------------
   function Is_Empty return Boolean is

   begin
      return Size = 0;
   end Is_Empty;

   ------------------------
   -- Timed_Event_In_Pos --
   ------------------------
   --
   --  This function can be used to get all the elements in the queue just
   --  incrementing 'Pos' form 1. The last element in the queue is detected
   --  because either 'Pos' reach Intex'Last or 'null' is returned. The
   --  order in which the elements are got when varying 'Pos' form 1 to
   --  Index'Last is not preorder, nor inorder, nor postorder.
   function Timed_Event_In_Pos (Pos : in Index) return Timed_Event_Ac is
   begin
      if Pos > Size then
         return null;
      else
         return Q (Pos);
      end if;
   end Timed_Event_In_Pos;
   pragma Inline (Timed_Event_In_Pos);

   -------------------------
   -- Show_TEQ_On_Console --
   -------------------------
   --
   --  Used when debugging
   procedure Show_TEQ_On_Console (Msg : in String) is
      use Basic_Console_IO;
   begin
      if Debug_Messages.Timed_Events_Queue_Debug_Messages then
         Put ("  |  ");
         Put (Msg);
         if Size = 0 then
            Put ("empty");
         else
            for I in 1..Size loop
               Put ("["); Kernel.Debug.Show_In_Secs (Q (I).T); Put ("]");
            end loop;
         end if;
      end if;
   end Show_TEQ_On_Console;


begin
   ------------
   --  Initialize Events Queue
   ----
   Size := 0;
   for I in Index loop
      Q (I) := null;
   end loop;
end Kernel.Timed_Events_Queue;
