------------------------------------------------------------------------------
--  ------------------         M a R T E   O S         -------------------  --
------------------------------------------------------------------------------
--                                                             V1.57  Dic 2005
--
--                   'S t a c k s _ M a n a g e m e n t'
--
--                                 Body
--
--
--  File 'stacks_management.adb'                                       By Mar.
--
--
--  Task stacks management. This package needs to be improved.
--
--
--  Stack_Pool_Base_Address -----+         Stack_Pool
--                               |    +-----------------+
--  Stack_List(1).Array_Integer -+--> | MAGIC_LABEL_EOS | <--- Top of stack 1
--                                    |                 |
--                                    |                 |
--                                    |                 |   +- Base of stack 1
--                                    +-----------------+   |
--  Stack_List(2).Array_Integer ----> | MAGIC_LABEL_EOS | <-+- Top of stack 2
--                                    |                 |
--                                           ...
--
--  ----------------------------------------------------------------------
--   Copyright (C) 2004   Universidad de Cantabria, SPAIN
--
--   MaRTE OS web page: http://marte.unican.es
--   Contact Addresses: Mario Aldea Rivas          aldeam@unican.es
--                      Michael González Harbour      mgh@unican.es
--
--  MaRTE OS  is free software; you can  redistribute it and/or  modify it
--  under the terms of the GNU General Public License  as published by the
--  Free Software Foundation;  either  version 2, or (at  your option) any
--  later version.
--
--  MaRTE OS  is distributed  in the  hope  that  it will be   useful, but
--  WITHOUT  ANY  WARRANTY;     without  even the   implied   warranty  of
--  MERCHANTABILITY  or  FITNESS FOR A  PARTICULAR PURPOSE.    See the GNU
--  General Public License for more details.
--
--  You should have received  a  copy of  the  GNU General Public  License
--  distributed with MaRTE  OS;  see file COPYING.   If not,  write to the
--  Free Software  Foundation,  59 Temple Place  -  Suite 330,  Boston, MA
--  02111-1307, USA.
--
--  As a  special exception, if you  link this  unit  with other  files to
--  produce an   executable,   this unit  does  not  by  itself cause  the
--  resulting executable to be covered by the  GNU General Public License.
--  This exception does  not however invalidate  any other reasons why the
--  executable file might be covered by the GNU Public License.
--
------------------------------------------------------------------------------
with System; -- for  'Address'


with Configuration_Parameters; -- for 'Num_Tasks_Mx'

with Kernel_Console;

package body Stacks_Management is

   ---------------------------------
   -- Magic Label at end of stack --
   ---------------------------------
   --
   --  This "Magic" number is writen at the end of every stack. A "Stack
   --  Overflow" error can be detected when this number is overwriten (with
   --  a different value of course ;-)
   MAGIC_LABEL_EOS : constant Integer := 16#60230e19#;

   --------------------------
   -- Stack points of view --
   --------------------------
   type Stack is array (DWord_Pos) of Integer;
   type Stack_Address is array (DWord_Pos) of System.Address;


   ---------------------------------
   -- 'Stack_Data' y 'Stack_List' --
   ---------------------------------
   type Access_Array_Integer is access all Stack;
   type Access_Array_Address is access all Stack_Address;
   function Access_Integer_To_Access_Address (Dat : in  Access_Array_Integer)
                                              return Access_Array_Address;
   pragma Import (C, Access_Integer_To_Access_Address,
                   "convert_4_bytes_long_data");
   function Address_To_Unsigned_32 (Add : in System.Address)
                                    return Unsigned_32;
   pragma Import (C, Address_To_Unsigned_32, "convert_4_bytes_long_data");
   --  Defined in 'miscellany.c'

   -----------------
   -- Stack_Data --
   -----------------
   type Stack_Data is record
      ID                   : Integer;
      Free                 : Boolean;
      Array_Integer        : Access_Array_Integer; -- This and this ...
      Array_Address        : Access_Array_Address; -- point to the same place
      Stack_Size_In_Bytes  : Integer;
      Stack_Size_In_DWords : Integer;
   end record;

   Stack_List : array (1 .. Configuration_Parameters.Num_MaRTE_Tasks_Mx)
     of aliased Stack_Data;

   --------------------
   -- Pool of stacks --
   --------------------
   Stack_Pool : array (1 .. Configuration_Parameters.Num_MaRTE_Tasks_Mx)
     of aliased Stack;

   -----------------------------
   -- Stack_Pool_Base_Address --
   -----------------------------
   --
   --  Base addres of the stack pool:
   --  Stack_Pool_Base_Address := Stack_Pool'Address
   Stack_Pool_Base_Address : Unsigned_32;

   -------------------
   -- Request_Stack --
   -------------------
   function Request_Stack (Size : in Integer) return Stack_ID is
   begin
      for I in 1 .. Configuration_Parameters.Num_MaRTE_Tasks_Mx loop
         if Stack_List (I).Free then
            Stack_List (I).Free := False;
            return Stack_List (I)'Access;
         end if;
      end loop;
      --  No free stacks!!
      Kernel_Console.Error ("#80#");
      return null;
   end Request_Stack;

   -------------------
   -- Release_Stack --
   -------------------
   procedure Release_Stack (The_Stack : in Stack_ID) is
   begin
      The_Stack.Free := True;
   end Release_Stack;

   ------------------------
   -- Stack_Base_Address --
   ------------------------
   --  XXX Only works if all the stacks have the same size
   function Stack_Base_Address (The_Stack : in Stack_ID)
                                return Unsigned_32 is
   begin
      return Stack_Pool_Base_Address +
        Unsigned_32 (The_Stack.ID * Stack_Size_In_Bytes);
   end Stack_Base_Address;

   -----------------------
   -- Stack_Top_Address --
   -----------------------
   --  XXX Only works if all the stacks have the same size
   function Stack_Top_Address (The_Stack : in Stack_ID)
                               return Unsigned_32 is
   begin
      return Stack_Pool_Base_Address +
        Unsigned_32 ((The_Stack.ID - 1) * Stack_Size_In_Bytes);
   end Stack_Top_Address;

   ----------------------------
   -- DWord_In_Stack_Address --
   ----------------------------
   --  XXX Only works if all the stacks have the same size
   function DWord_In_Stack_Address (The_Pos      : in DWord_Pos;
                                    In_The_Stack : in Stack_ID)
                                    return Unsigned_32 is
   begin
      return Stack_Base_Address (In_The_Stack) - Unsigned_32 (The_Pos) * 4;
   end DWord_In_Stack_Address;

   -----------------------------
   -- Get_Stack_Size_In_Bytes --
   -----------------------------
   function Get_Stack_Size_In_Bytes (The_Stack : in Stack_ID)
                                     return Integer is
   begin
      return The_Stack.Stack_Size_In_Bytes;
   end Get_Stack_Size_In_Bytes;

   -------------------------------------
   -- Write_In_Stack (System.Address) --
   -------------------------------------
   procedure Write_In_Stack (Addr       : in System.Address;
                             In_The_Pos : in DWord_Pos;
                             In_Stack   : in Stack_ID) is
   begin
      In_Stack.Array_Address (DWord_Pos (In_Stack.Stack_Size_In_DWords)
                              - In_The_Pos + 1) := Addr;
   end Write_In_Stack;

   ------------------------------
   -- Write_In_Stack (Integer) --
   ------------------------------
   procedure Write_In_Stack (Int        : in Integer;
                             In_The_Pos : in DWord_Pos;
                             In_Stack   : in Stack_ID) is
   begin
      In_Stack.Array_Integer (DWord_Pos (In_Stack.Stack_Size_In_DWords)
                              - In_The_Pos + 1) := Int;
   end Write_In_Stack;

   ---------------------
   -- Read_From_Stack --
   ---------------------
   function Read_From_Stack (Pos   : in DWord_Pos;
                             Stack : in Stack_ID) return Integer is
   begin
      return Stack.Array_Integer (DWord_Pos (Stack.Stack_Size_In_DWords)
                                  - Pos + 1);
   end Read_From_Stack;

   --------------------
   -- Stack_Overflow --
   --------------------
   function Stack_Overflow (Stack : in Stack_ID) return Boolean is
   begin
      return Stack.Array_Integer (DWord_Pos'First) /= MAGIC_LABEL_EOS;
   end Stack_Overflow;


begin
   Stack_Pool_Base_Address :=
     Address_To_Unsigned_32 (Stack_Pool'Address);
   for I in 1 .. Configuration_Parameters.Num_MaRTE_Tasks_Mx loop
      Stack_List (I).ID    := I;
      Stack_List (I).Free  := True;
      Stack_List (I).Array_Integer := Stack_Pool (I)'Access;
      Stack_List (I).Array_Address :=
        Access_Integer_To_Access_Address (Stack_List (I).Array_Integer);
      Stack_List (I).Stack_Size_In_Bytes  := Stack_Size_In_Bytes;
      Stack_List (I).Stack_Size_In_DWords := Stack_Size_In_DWords;
      Stack_List (I).Array_Integer (DWord_Pos'First) := MAGIC_LABEL_EOS;
   end loop;
end Stacks_Management;
