pragma Source_Reference (1, "hardware_interface.gps");
------------------------------------------------------------------------------
--  ------------------         M a R T E   O S         -------------------  --
------------------------------------------------------------------------------
--                                                             V1.57  Dic 2005
--
--                   'H a r d w a r e _ I n t e r f a c e'
--
--                                  Spec
--
--
--  File 'hardware_interface.ads'                By MAR and
--                                                  Miguel ngel Masmano Tello.
--
--  Linux Version.
--
--  This package defines the interface between the hardware dependent
--  and hardware independent parts of the kernel.
--
--  The idea of running MaRTE OS as a Linux process is due to Miguel ngel
--  Masmano     Tello     (Universidad    Politcnica     de     Valencia)
--  <mimastel@doctor.upv.es>.
--  He  is  also  the  author  of   most  of  the  code  involved  in  the
--  implementation of the hardware interface for this architecture.
--
--  ----------------------------------------------------------------------
--   Copyright (C) 2004   Universidad de Cantabria, SPAIN
--
--   MaRTE OS web page: http://marte.unican.es
--   Contact Addresses: Mario Aldea Rivas          aldeam@unican.es
--                      Michael González Harbour      mgh@unican.es
--
--  MaRTE OS  is free software; you can  redistribute it and/or  modify it
--  under the terms of the GNU General Public License  as published by the
--  Free Software Foundation;  either  version 2, or (at  your option) any
--  later version.
--
--  MaRTE OS  is distributed  in the  hope  that  it will be   useful, but
--  WITHOUT  ANY  WARRANTY;     without  even the   implied   warranty  of
--  MERCHANTABILITY  or  FITNESS FOR A  PARTICULAR PURPOSE.    See the GNU
--  General Public License for more details.
--
--  You should have received  a  copy of  the  GNU General Public  License
--  distributed with MaRTE  OS;  see file COPYING.   If not,  write to the
--  Free Software  Foundation,  59 Temple Place  -  Suite 330,  Boston, MA
--  02111-1307, USA.
--
--  As a  special exception, if you  link this  unit  with other  files to
--  produce an   executable,   this unit  does  not  by  itself cause  the
--  resulting executable to be covered by the  GNU General Public License.
--  This exception does  not however invalidate  any other reasons why the
--  executable file might be covered by the GNU Public License.
--
-------------------------------------------------------------------------------
with System;
with Basic_Integer_Types; use Basic_Integer_Types;
with Interrupt_Tables;
pragma Elaborate_All (Interrupt_Tables);
with Processor_Registers;

package Hardware_Interface is

   pragma Elaborate_Body;

   ----------------
   -- Interrupts --
   ----------------
   subtype HW_Interrupt is Interrupt_Tables.IRQ;

   TIMER_INTERRUPT : constant HW_Interrupt := Interrupt_Tables.TIMER_IRQ;

   subtype Trap_State is Interrupt_Tables.Trap_State; -- PC
   subtype Trap_State_Ac is Interrupt_Tables.Trap_State_Ac;

   subtype HW_Interrupt_Handler is
     Interrupt_Tables.IRQ_Handler_Type;

   procedure Default_HW_Interrupt_Handler (State : in Trap_State_Ac);

   procedure Install_HW_Interrupt_Handler
     (Int_Num  : in HW_Interrupt;
      Handler  : in HW_Interrupt_Handler);


   procedure Interrupts_Disabled;
   pragma Export (C, Interrupts_Disabled, "hwi_disable_interrupts");

   procedure Interrupts_Enabled;
   pragma Import (C, Interrupts_Enabled,
                    "linux_signals_unmask_all_signals");
   --  Defined in 'boot/linux_signals.c'

   procedure Interrupt_Controller_Hardware_Enable_Interrupt
     (Int : in HW_Interrupt);
   pragma Import (C, Interrupt_Controller_Hardware_Enable_Interrupt,
                    "linux_signals_add_signal_to_used_set");
   --  Defined in 'boot/linux_signals.c'

   procedure Interrupt_Controller_Hardware_Disable_Interrupt
     (Int : in HW_Interrupt);
   pragma Import (C, Interrupt_Controller_Hardware_Disable_Interrupt,
                    "linux_signals_del_signal_from_used_set");
   --  Defined in 'boot/linux_signals.c'

   procedure Interrupt_Controller_Hardware_Enabled;
   pragma Inline (Interrupt_Controller_Hardware_Enabled);
   --  It doesn't do anything in this architecture

   function Are_Interrupts_Enabled return Boolean;
   pragma Inline (Are_Interrupts_Enabled);


   -----------------------------------------
   -- Save and Restore the Flags Register --
   -----------------------------------------
   procedure Save_Flags_And_Disable_Interrupts (Not_Used : out Integer);

   function Save_Flags return Integer;
   pragma Inline (Save_Flags);

   procedure Restore_Flags (Not_Used : in Integer);


   ----------------------------
   -- Stack Pointer Register --
   ----------------------------
   function Get_Stack_Pointer_Register return Unsigned_32
     renames Processor_Registers.Get_ESP_Register;
   pragma Inline (Get_Stack_Pointer_Register);


   --------------------
   -- Bit Operations --
   --------------------
   procedure Bit_Set (Bit_Field : in out Unsigned_32; Bit : in Unsigned_32)
     renames Processor_Registers.Bit_Set;
   pragma Inline (Bit_Set);
   --  Sets to 1 the bit in 'Bit_Field' in the position pointed by 'Bit'
   --  ('Bit' in range 0 .. 31)

   procedure Bit_Reset (Bit_Field : in out Unsigned_32; Bit : in Unsigned_32)
     renames Processor_Registers.Bit_Reset;
   pragma Inline (Bit_Reset);
   --  Sets to 0 the bit in 'Bit_Field' in the position pointed by 'Bit'
   --  ('Bit' in range 0 .. 31)

   procedure Bit_Scan_Forward (Bit_Field : in Unsigned_32;
                               Bit       : out Unsigned_32)
     renames Processor_Registers.Bit_Scan_Forward; -- 'Bit_Field' can't be 0!!
   pragma Inline (Bit_Scan_Forward);
   --  Returns in 'Bit' the position of the LEST SIGNIFICANT bit set in
   --  'Bit_Field' ('Bit' in range 0 .. 31)

   procedure Bit_Scan_Reverse (Bit_Field : in Unsigned_32;
                               Bit       : out Unsigned_32)
     renames Processor_Registers.Bit_Scan_Reverse; -- 'Bit_Field' can't be 0!!
   pragma Inline (Bit_Scan_Reverse);
   --  Returns in 'Bit' the position of the MOST SIGNIFICANT bit set in
   --  'Bit_Field' ('Bit' in range 0 .. 31)

   ---------------------
   -- Time and Timers --
   ---------------------

   Last_HW_Timer : constant Integer_32 := 0; -- PC

   type HW_Timers is new Integer_32 range 0 .. Last_HW_Timer;
   for HW_Timers'Size use 32;

   HW_Timer_0 : constant HW_Timers := 0;

   type HWTime is new Integer_64;

   function Get_HWTime_Slow return HWTime;
   pragma Inline (Get_HWTime_Slow);

   function Get_HWTime return HWTime;
   pragma Inline (Get_HWTime);


   function Safe_Longest_Timer_Interval return HWTime;
   --  The longest interval that can be used to program the timer.


   function HWTime_To_Duration (Th : in HWTime) return Duration;
   pragma Inline (HWTime_To_Duration);

   function Duration_To_HWTime (D : in Duration)  return HWTime;
   pragma Inline (Duration_To_HWTime);


   function Get_HWClock_Frequency return HWTime;
   --  Ticks per second
   pragma Inline (Get_HWClock_Frequency);

   procedure Program_Timer (Timer           : in HW_Timers;
                            Interval        : in HWTime;
                            Next_Activation : out HWTime);
   pragma Inline (Program_Timer);

   function Compulsory_Timer_Reprogramming return Boolean;
   --  Returns true if the timer should always be reprogrammed even if
   --  there isn't any event. This could happen, for example, when the
   --  timer interrupts are also used for the system clock.

   ---------------------
   -- Real Time Clock --
   ---------------------
   RTC_Available : constant Boolean := True;
   function RTC_HWTime_Since_Epoch return HWTime;

   -------------------
   -- Context Swich --
   -------------------
   procedure Context_Switch (Old_Task : in System.Address;
                             New_Task : in System.Address)
     renames Processor_Registers.Context_Switch;
   pragma Inline (Context_Switch);

   procedure Change_Return_Address_Of_Preempted_Task
     (Top_Of_Stack    : in Unsigned_32;
      New_Ret_Address : in System.Address)
     renames Processor_Registers.Change_Return_Address_Of_Preempted_Task;

   -------------------------
   -- Finish_Hardware_Use --
   -------------------------
   procedure Finish_Hardware_Use;
   --  Invoked while finishing the application to leave the hardware used
   --  by MaRTE OS in a known and stable status. Called from 'our_exit' in
   --  'base_console_init.c'.



end Hardware_Interface;
