/*----------------------------------------------------------------------------
 *-------------------------      M a R T E   O S      ------------------------
 *----------------------------------------------------------------------------
 *                                                             V1.57  Dic 2005
 *                                'm a l l o c'
 *
 *                                      C
 *
 * File 'malloc.c'                                   By Miguel Masmano Tello
 *                                                      and MAR.
 *
 * Linux Library version of this file.
 *
 * Implements two versions of malloc:
 *
 * - Two Levels Segregate Fit memory allocator (TLSF): This algorithm
 *   allows memory allocation and deallocation with bounded response
 *   times (Developed at the Universidad Politcnica de Valencia by
 *   Miguel Masmano Tello <mmasmano@disca.upv.es>)
 *
 * - Simple algorithm: allows allocation of memory blocks with not
 *   deallocation. It is useful for applications that do not uses
 *   dynamic memory at all or uses it only at initialization time.
 *
 * ---------------------------------------------------------------------------
 *
 * Two Levels Segregate Fit memory allocator (TLSF)
 * Version 1.1
 *
 * Written by Miguel Masmano Tello <mmasmano@disca.upv.es>
 * Copyright (C) Dec, 2002 OCERA Consortium
 * Release under the terms of the GNU General Public License Version 2
 *
 * ----------------------------------------------------------------------
 *  Copyright (C) 2004   Universidad de Cantabria, SPAIN
 *
 *  MaRTE OS web page: http://marte.unican.es
 *  Contact Addresses: Mario Aldea Rivas          aldeam@unican.es
 *                     Michael González Harbour      mgh@unican.es
 *
 * MaRTE OS  is free software; you can  redistribute it and/or  modify it
 * under the terms of the GNU General Public License  as published by the
 * Free Software Foundation;  either  version 2, or (at  your option) any
 * later version.
 *
 * MaRTE OS  is distributed  in the  hope  that  it will be   useful, but
 * WITHOUT  ANY  WARRANTY;     without  even the   implied   warranty  of
 * MERCHANTABILITY  or  FITNESS FOR A  PARTICULAR PURPOSE.    See the GNU
 * General Public License for more details.
 *
 * You should have received  a  copy of  the  GNU General Public  License
 * distributed with MaRTE  OS;  see file COPYING.   If not,  write to the
 * Free Software  Foundation,  59 Temple Place  -  Suite 330,  Boston, MA
 * 02111-1307, USA.
 *
 * As a  special exception, if you  link this  unit  with other  files to
 * produce an   executable,   this unit  does  not  by  itself cause  the
 * resulting executable to be covered by the  GNU General Public License.
 * This exception does  not however invalidate  any other reasons why the
 * executable file might be covered by the GNU Public License.
 *
 *---------------------------------------------------------------------------*/
#include <sys/marte_configuration_parameters.h>

#ifdef USE_TLSF_MEMORY_ALLOCATOR
/*
 * Two Levels Segregate Fit memory allocator (TLSF)
 * Version 1.1
 *
 * Written by Miguel Masmano Tello <mmasmano@disca.upv.es>
 * Copyright (C) Dec, 2002 OCERA Consortium
 * Release under the terms of the GNU General Public License Version 2
 *
 */
#include <unistd.h>
#include <misc/assert.h>
#include "tlsf.h"
#include "tlsf_mapping.h"

/* see man malloc */
/*
 * malloc searchs a free block of size 'size',
 * after the free block will be splitted in two new blocks,
 * one of these new blocks will be given to the user and the
 * other will be inserted into TLSF structure
 *
 * The cost of this operation is
 *      best case: (K) = (1)
 *      worst case: (MAX_FL_LOG2_INDEX - MIN_FL_LOG2_INDEX + MAX_SL_INDEX + K)
 *                   = (1)
 * where K is a constant integer
 */

void *marte_malloc (size_t size)
{
  // << MaRTE
  // TLSF_t *ptr_TLSF;
  // MaRTE >>
  int flags, fl, sl, n, found = 0;
  int old_size = BYTES2TLSF_WORDS(size), last_block;
  block_header_t *bh = NULL, *bh2 = NULL, *bh3 = NULL;

  // << MaRTE
  // ptr_TLSF = (TLSF_t *) block_ptr;
  // MaRTE >>

  if (ptr_TLSF == NULL || ptr_TLSF -> magic_number != MAGIC_NUMBER) {
    PRINT_DBG_C ("FATAL ERROR: TLSF structure not initialized\n");
    return NULL;
  }
    
  if (!(size > 0)) {    
    // << MaRTE
    // PRINT_DBG_C ("ERROR: Memory pool exhausted!!!\n");
    PRINT_DBG_C ("ERROR: called malloc with size < 0!!!\n");
    // MaRTE >>
    return NULL;
  }

  if (old_size < MIN_SIZE) {
    size = MIN_SIZE;
    fl = 0;
    sl = 0;
  } else {
    mapping_function (old_size, &fl, &sl, ptr_TLSF);

    if (++sl == ptr_TLSF -> max_sl_index) {
      fl ++;
      sl = 0;
    }
    
    /*
     * This is the reason of the internal fragmentation
     * The block given is greater that the size demanded
     */
    
    /* size can be smaller that maximum SLI, in this case the mapping 
       function has problems calculating fl and sl values */
    
    if (old_size >= 
	BYTES2TLSF_WORDS(ptr_TLSF -> max_sl_index << 2)) {
      size = (1 << fl);
      size += (sl << (fl - ptr_TLSF -> max_sl_log2_index));
    } else {
      size = old_size + 1;
      sl = old_size - MIN_SIZE;
    }
  
    fl -= MIN_LOG2_SIZE;
  }

  /*----------------------------------------*/
  /* The search for a free block begins now */
  /*----------------------------------------*/
  
  /*
   * Our first try, we take the first free block 
   * from fl_array or its buddy
   */
  
  flags = THREAD_LOCK;

  sl = ptr_TLSF -> fl_array[fl].bitmapSL & ((~0) << sl);
  if (sl != 0) {
    // << MaRTE
    // sl = TLSF_fls(sl);
    bit_scan_reverse (sl, &sl);
    // MaRTE >>
    bh =  ptr_TLSF -> fl_array [fl].sl_array [sl];
    ptr_TLSF -> fl_array [fl].sl_array [sl] = bh -> ptr.free_ptr.next;
    if (ptr_TLSF -> fl_array [fl].sl_array [sl] != NULL)
      ptr_TLSF ->fl_array [fl].sl_array [sl] -> ptr.free_ptr.prev = NULL;
    else {
      TLSF__clear_bit (sl, ptr_TLSF -> fl_array[fl].bitmapSL);
      if (!ptr_TLSF -> fl_array[fl].bitmapSL)
        TLSF__clear_bit (fl, ptr_TLSF -> bitmapFL);
    }
    found = 1;
    goto out;
  }
  
  /*
   * On the last case a free block is searched using a bitmap
   */
  
  // << MaRTE
  // fl = TLSF_fls(ptr_TLSF -> bitmapFL & ((~0) << (fl + 1)));
  bit_scan_reverse (ptr_TLSF -> bitmapFL & ((~0) << (fl + 1)), &fl);
  // MaRTE >>
  
  if (fl > 0) {  
    // << MaRTE
    // sl = TLSF_fls(ptr_TLSF -> fl_array[fl].bitmapSL);
    bit_scan_reverse (ptr_TLSF -> fl_array[fl].bitmapSL, &sl);
    // MaRTE >>
    bh = ptr_TLSF -> fl_array [fl].sl_array [sl];
    ptr_TLSF -> fl_array [fl].sl_array [sl] = bh -> ptr.free_ptr.next;
    if (ptr_TLSF -> fl_array [fl].sl_array [sl] != NULL){
      ptr_TLSF -> fl_array [fl].sl_array [sl] 
	-> ptr.free_ptr.prev = NULL;
    } else {
      TLSF__clear_bit (sl, ptr_TLSF -> fl_array[fl].bitmapSL);
      if (!ptr_TLSF -> fl_array[fl].bitmapSL)
        TLSF__clear_bit (fl, ptr_TLSF -> bitmapFL);
    }
    found = 1;
    goto out;
  }
  /* end of the search */
  /*------------------------------------------------------------*/
  
 out:
  
  /*
   * HUGGGG, NOT ENOUGHT MEMORY
   * I think that we have done all that we have been able, I'm sorry
   */
  
  if (!found) {
    THREAD_UNLOCK(flags);
    PRINT_DBG_C ("ERROR: Memory pool exhausted!!!\n");
    return NULL;
  }
  
  /*
   * we can say: YESSSSSSSSSSS, we have enought memory!!!!
   */
  
  /* can bh be splitted? */ 
  
  n = (int)(GET_BLOCK_SIZE(bh) - size - beg_header_overhead);
  if (n >= (int) MIN_SIZE) {
    /*
     * Yes, bh will be splitted
     */

    /* The new block will begin at the end of the current block */
    last_block = IS_LAST_BLOCK(bh)?1:0;
    bh -> size = size;
    SET_USED_BLOCK(bh);
    
    bh2 = (block_header_t *) (__u8 *) (bh -> ptr.buffer + 
				       TLSF_WORDS2BYTES
				       (GET_BLOCK_SIZE(bh)));
    bh2 -> prev_phys_block = bh;
    bh2 -> size = n;
    if (last_block) SET_LAST_BLOCK (bh2);
    mapping_function (GET_BLOCK_SIZE(bh2), &fl, &sl, ptr_TLSF);
    fl -= MIN_LOG2_SIZE;
    bh2 -> ptr.free_ptr.first_index = fl;
    bh2 -> ptr.free_ptr.second_index = sl;
    bh2 -> ptr.free_ptr.prev = NULL;
    bh2 -> ptr.free_ptr.next = ptr_TLSF -> fl_array [fl].sl_array [sl];
    
    if (!IS_LAST_BLOCK (bh2)) {
      bh3 = (block_header_t *) (__u8 *) (bh2 -> ptr.buffer + 
					 TLSF_WORDS2BYTES
					 (GET_BLOCK_SIZE(bh2)));
      bh3 -> prev_phys_block = bh2;
    }
    

    if (ptr_TLSF -> fl_array [fl].sl_array [sl] != NULL)
      ptr_TLSF -> fl_array [fl].sl_array [sl] -> ptr.free_ptr.prev = bh2;
    ptr_TLSF -> fl_array [fl].sl_array [sl] = bh2;
    TLSF__set_bit (sl, ptr_TLSF -> fl_array[fl].bitmapSL);
    TLSF__set_bit (fl, ptr_TLSF -> bitmapFL); 
  }
  
  SET_USED_BLOCK(bh);
  // << MaRTE
  assert(((char *)(bh -> ptr.buffer) > (char *)ptr_TLSF) && 
	 ((char *)(bh -> ptr.buffer) < 
	  ((char *)ptr_TLSF + DYNAMIC_MEMORY_POOL_SIZE_IN_K * 1024)));
  // MaRTE >>

  THREAD_UNLOCK(flags);
  
  return (void *)  bh -> ptr.buffer; 
}

#else // USE_TLSF_MEMORY_ALLOCATOR
#include <sys/types.h>

/*
 * Symple dynamic memory algorithm
 */
extern void *simple_dynamic_memory__malloc (size_t size);

void *marte_malloc (size_t size)
{
  return simple_dynamic_memory__malloc (size);
}

#endif // USE_TLSF_MEMORY_ALLOCATOR
