------------------------------------------------------------------------------
-- --------------------        M a R T E   O S          ------------------- --
------------------------------------------------------------------------------
--                                                             V1.57  Dic 2005
--
--                       'D r i v e r s _ M a R T E'
--
--                                   Spec
--
--
--  File 'drivers_marte.ads'                                     By Fguerreira
--
--
--  General types and functions that any driver may need
--
--  ----------------------------------------------------------------------
--   Copyright (C) 2004   Universidad de Cantabria, SPAIN
--
--   MaRTE OS web page: http://marte.unican.es
--   Contact Addresses: Mario Aldea Rivas          aldeam@unican.es
--                      Michael González Harbour      mgh@unican.es
--
--  MaRTE OS  is free software; you can  redistribute it and/or  modify it
--  under the terms of the GNU General Public License  as published by the
--  Free Software Foundation;  either  version 2, or (at  your option) any
--  later version.
--
--  MaRTE OS  is distributed  in the  hope  that  it will be   useful, but
--  WITHOUT  ANY  WARRANTY;     without  even the   implied   warranty  of
--  MERCHANTABILITY  or  FITNESS FOR A  PARTICULAR PURPOSE.    See the GNU
--  General Public License for more details.
--
--  You should have received  a  copy of  the  GNU General Public  License
--  distributed with MaRTE  OS;  see file COPYING.   If not,  write to the
--  Free Software  Foundation,  59 Temple Place  -  Suite 330,  Boston, MA
--  02111-1307, USA.
--
--  As a  special exception, if you  link this  unit  with other  files to
--  produce an   executable,   this unit  does  not  by  itself cause  the
--  resulting executable to be covered by the  GNU General Public License.
--  This exception does  not however invalidate  any other reasons why the
--  executable file might be covered by the GNU Public License.
--
------------------------------------------------------------------------------

with Kernel;
with Kernel.File_System_Data_Types;
with Basic_Integer_Types;
with Kernel.Mutexes;
with Kernel.Tasks_Operations.Internals;


package Drivers_MaRTE is

   subtype Int is Basic_Integer_Types.Int;
   subtype Ssize_T is Basic_Integer_Types.Ssize_T;

   subtype File_Descriptor   is Kernel.File_System_Data_Types.File_Descriptor;
   subtype File_Access_Mode  is Kernel.File_System_Data_Types.File_Access_Mode;
   subtype Major             is Kernel.File_System_Data_Types.Major;
   subtype Minor             is Kernel.File_System_Data_Types.Minor;
   subtype Buffer_Length     is Kernel.File_System_Data_Types.Buffer_Length;
   subtype Buffer_Ac         is Kernel.File_System_Data_Types.Buffer_Ac;
   subtype Ioctl_Option_Value is
     Kernel.File_System_Data_Types.Ioctl_Option_Value;

   ----------------------------------------------------------------------------
   --  Drivers Functions Specifications ---------------------------------------
   ----------------------------------------------------------------------------
   subtype Create_Function_Ac is
     Kernel.File_System_Data_Types.Create_Function_Ac;
     --  function Create return Int;

   subtype Remove_Function_Ac is
     Kernel.File_System_Data_Types.Remove_Function_Ac;
     --  function Remove return Int;

   subtype Open_Function_Ac is
     Kernel.File_System_Data_Types.Open_Function_Ac;
     --  function Open (Fd   : in File_Descriptor;
     --                 Mode : in File_Access_Mode) return Int;

   subtype Close_Function_Ac is
     Kernel.File_System_Data_Types.Close_Function_Ac;
     --  function Close (Fd : in File_Descriptor) return Int;

   subtype Read_Function_Ac is
     Kernel.File_System_Data_Types.Read_Function_Ac;
     --  function Read (Fd         : in File_Descriptor;
     --                 Buffer_Ptr : in Buffer_Ac;
     --                 Bytes      : in Buffer_Length) return Ssize_T;

   subtype Write_Function_Ac is
     Kernel.File_System_Data_Types.Write_Function_Ac;
     --  function Write (Fd         : in File_Descriptor;
     --                  Buffer_Ptr : in Buffer_Ac;
     --                  Bytes      : in Buffer_Length) return Ssize_T;

   subtype Ioctl_Function_Ac is
     Kernel.File_System_Data_Types.Ioctl_Function_Ac;
     --  function Ioctl (Fd             : in File_Descriptor;
     --                  Request        : in Ioctl_Option_Value;
     --                  Ioctl_Data_Ptr : in Buffer_Ac) return Int;

   ----------------------------------------------------------------------------
   --  Operations available to drivers ----------------------------------------
   ----------------------------------------------------------------------------

   --  Gets the major number associated to a file descriptor
   procedure Get_Major_Number (Fd         : in  File_Descriptor;
                               Mj         : out Major;
                               Invalid_Fd : out Boolean);
   --  C version 'get_major' (defined in the package body) returns -1
   --  for an invalid file descriptor

   --  Gets the minor number associated to a file descriptor
   procedure Get_Minor_Number (Fd         : in  File_Descriptor;
                               Mn         : out Minor;
                               Invalid_Fd : out Boolean);
   --  C version 'get_minor' (defined in the package body) returns -1
   --  for an invalid file descriptor


   --  Sets POSIX error for a driver operation
   procedure Set_POSIX_Error (Error : in Kernel.Error_Code) renames
     Kernel.Tasks_Operations.Internals.Set_POSIX_Error;

   --  Gets current POSIX error
   function Get_POSIX_Error return Kernel.Error_Code renames
     Kernel.Tasks_Operations.Internals.Get_Last_POSIX_Error_Code;


   --  Gets specific data associated to a file descriptor. This
   --  specific data field (an integer) can be converted into an
   --  access type using an unchecked conversion.
   procedure Get_Specific_Data (Fd         : in  File_Descriptor;
                                Data       : out Integer;
                                Invalid_Fd : out Boolean);
   --  C version 'Driver_GetSpecific' (defined in the package body) returns -1
   --  for an invalid file descriptor

   --  Sets specific data associated to a file descriptor
   procedure Set_Specific_Data (Fd         : in  File_Descriptor;
                                Data       : in  Integer;
                                Invalid_Fd : out Boolean);
   --  C version 'Driver_SetSpecific' (defined in the package body) returns -1
   --  for an invalid file descriptor



--     function "<" (A : in Buffer_Length; B : in Buffer_Length)
--                  return Boolean
--       renames Kernel.File_System_Data_Types."<";
--     function "<=" (A : in Buffer_Length; B : in Buffer_Length)
--                   return Boolean
--       renames Kernel.File_System_Data_Types."<=";
--     function ">" (A : in Buffer_Length; B : in Buffer_Length)
--                  return Boolean
--       renames Kernel.File_System_Data_Types.">";
--     function ">=" (A : in Buffer_Length; B : in Buffer_Length)
--                   return Boolean
--       renames Kernel.File_System_Data_Types.">=";
--     function "=" (A : in Buffer_Length; B : in Buffer_Length)
--                  return Boolean
--       renames Kernel.File_System_Data_Types."=";


end Drivers_MaRTE;

