/*----------------------------------------------------------------------------
 *-------------------------      M a R T E   O S      ------------------------
 *----------------------------------------------------------------------------
 *                                                             V1.57  Dic 2005
 *
 *                   's e r i a l _ p o r t _ d r i v e r _ c'
 *
 *                                      C
 *
 *  File 'serial_port_driver_c.c'                           by Fguerreira and
 *                                                             MAR.
 *
 *  I/O functions used by the serial port. Pooling mode, no interrupts
 *  supported.
 *
 *  If this driver is going to be used as stdout and/or stderr devices
 *  '/dev/stdout' and '/dev/stderr' ('k-devices_table.ads') must have
 *  assigned 0 as their minor numbers. Otherwise value of
 *  CONSOLE_SERIAL_PORT should be changed.
 *
 *
 * Based on code taken from the "Flux OSKit"
 * Copyright (c) 1996, 1998 University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the Flux OSKit.  The OSKit is free software, also known
 * as "open source;" you can redistribute it and/or modify it under the terms
 * of the GNU General Public License (GPL), version 2, as published by the Free
 * Software Foundation (FSF).  To explore alternate licensing terms, contact
 * the University of Utah at csl-dist@cs.utah.edu or +1-801-585-3271.
 * 
 * The OSKit is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GPL for more details.  You should have
 * received a copy of the GPL along with the OSKit; see the file COPYING.  If
 * not, write to the FSF, 59 Temple Place #330, Boston, MA 02111-1307, USA.
 *
 *
 *  ----------------------------------------------------------------------
 *   Copyright (C) 2004   Universidad de Cantabria, SPAIN
 *
 *   MaRTE OS web page: http://marte.unican.es
 *   Contact Addresses: Mario Aldea Rivas          aldeam@unican.es
 *                      Michael González Harbour      mgh@unican.es
 *
 *  MaRTE OS  is free software; you can  redistribute it and/or  modify it
 *  under the terms of the GNU General Public License  as published by the
 *  Free Software Foundation;  either  version 2, or (at  your option) any
 *  later version.
 *
 *  MaRTE OS  is distributed  in the  hope  that  it will be   useful, but
 *  WITHOUT  ANY  WARRANTY;     without  even the   implied   warranty  of
 *  MERCHANTABILITY  or  FITNESS FOR A  PARTICULAR PURPOSE.    See the GNU
 *  General Public License for more details.
 *
 *  You should have received  a  copy of  the  GNU General Public  License
 *  distributed with MaRTE  OS;  see file COPYING.   If not,  write to the
 *  Free Software  Foundation,  59 Temple Place  -  Suite 330,  Boston, MA
 *  02111-1307, USA.
 *
 *  As a  special exception, if you  link this  unit  with other  files to
 *  produce an   executable,   this unit  does  not  by  itself cause  the
 *  resulting executable to be covered by the  GNU General Public License.
 *  This exception does  not however invalidate  any other reasons why the
 *  executable file might be covered by the GNU Public License.
 *
 *---------------------------------------------------------------------------*/


#include <drivers/serial_port_driver.h>
#include <drivers/drivers_marte.h>  // For 'get_minor'

#include <unistd.h>
#include <stdio.h>
#include <sys/pio.h>


 /* Base directions for each port */
static int ser_io_base[4] = {0x3F8, 0x2F8, 0x3E8, 0x2E8};

/* Base directions for each port */
static termios_t ser_termios[4];

/* Forward functions declarations */
static void set_port_attr (int port, termios_t *com_params);
ssize_t basic_write (unsigned int port, void *buffer, size_t bytes);
int serial_port_ioctl (int file_descriptor, int request, void* argp);


/* Basic default termios settings for a typical raw-mode console. */

termios_t raw_termios =
  {
    0,             /* input flags */
    0,             /* output flags */
    CS8,           /* control flags */
    0,             /* local flags */
    {
      'D'-64,                   /* VEOF */
      VDISABLE,                 /* VEOL */
      0,
      'H'-64,                   /* VERASE */
      0,
      'U'-64,                   /* VKILL */
      0,
      0,
      'C'-64,                   /* VINTR */
      '\\'-64,                  /* VQUIT */
      'Z'-64,                   /* VSUSP */
      0,
      'Q'-64,                   /* VSTART */
      'S'-64,                   /* VSTOP */
    },
    B9600,         /* input speed */
    B9600,         /* output speed */
  };



/* Basic default termios settings for a typical cooked-mode console. */

termios_t cooked_termios =
  {
    ICRNL | IXON,                           /* input flags */
    OPOST,                                  /* output flags */
    CS8,                                    /* control flags */
    ECHO | ECHOE | ECHOK | ICANON,          /* local flags */
    {
      'D'-64,                   /* VEOF */
      VDISABLE,                 /* VEOL */
      0,
      'H'-64,                   /* VERASE */
      0,
      'U'-64,                   /* VKILL */
      0,
      0,
      'C'-64,                   /* VINTR */
      '\\'-64,                  /* VQUIT */
      'Z'-64,                   /* VSUSP */
      0,
      'Q'-64,                   /* VSTART */
      'S'-64,                   /* VSTOP */
    },
    B9600,         /* input speed */
    B9600,         /* output speed */
  };

/******************************/
/*  serial_port_init_console  */
/******************************/
// Configure the serial port to be used from the very beggining of
// excution (before 'open' has been invoked). This is necessary when
// serial port is going to be used as stdout and/or stderr.
#define CONSOLE_SERIAL_PORT 0 // In 'k-devices_table.ads',
	       // devices '/dev/stdout' and '/dev/stderr' must have
	       // assigned 0 as their minor numbers.
int serial_port_init_console ()
{
  static int initialized = 0;
  if (!initialized) {
    // Port is opened with RAW-MODE (8N1)
    // (8N1) -> (8 bits, No Parity bit, 1 Stop bit)
    ser_termios [CONSOLE_SERIAL_PORT] = raw_termios;
    // ispeed = ospeed = B115200
    ser_termios [CONSOLE_SERIAL_PORT].ispeed = B115200;
    ser_termios [CONSOLE_SERIAL_PORT].ospeed = B115200;
    // set port attributes
    set_port_attr (CONSOLE_SERIAL_PORT, &ser_termios [CONSOLE_SERIAL_PORT]);
    initialized = 1;
  }
  return 0;
}

/**********************************/
/*  End of kernel initialization  */
/**********************************/
void serial_port_end_of_kernel_initialization ()
{
}

/*******************************/
/*  serial_port_console_write  */
/*******************************/
// Used when serial port is going to be used as stdout and/or
// stderr.
ssize_t serial_port_console_write (int file_descriptor, void *buffer, 
				   size_t bytes)
{
  return basic_write (CONSOLE_SERIAL_PORT, buffer, bytes);
}

/************/
/*  Create  */
/************/
/* Not necessary
int serial_port_create ()
{
  // Initialization is performed in 'open'
  return 0;
} */


/************/
/*  Remove  */
/************/
/* Not necessary
int serial_port_remove ()
{
  return 0;
}*/



/**********/
/*  Open  */
/**********/

int serial_port_open (int file_descriptor, int file_access_mode)
{
  int port;

  port = get_minor (file_descriptor);

  /* Port is opened with RAW-MODE (8N1) as default */
  /* (8N1) -> (8 bits, No Parity bit, 1 Stop bit) */
  ser_termios [port] = raw_termios;
  serial_port_ioctl (file_descriptor, SETATTR, ((void *)&ser_termios [port]));
  return 0;
}



/***********/
/*  Close  */
/***********/
/* Not necessary
int serial_port_close (int file_descriptor)
{
  return 0;
}*/




/**********/
/*  Read  */
/**********/

ssize_t serial_port_read (int file_descriptor, void *buffer, size_t bytes)
{
  int i, port, byte, bytes_read = 0;
  unsigned char mode;

  port = get_minor (file_descriptor);

  for(i=0; i<bytes; i++)
    {

      /* Wait for a character to arrive.  */
      for (;;)
	{
	  /* Grab it.  */
	  mode = inb_p(ser_io_base[port] + 3);
          outb_p(ser_io_base[port] + 3, 0x7F & mode);     /* DLAB = 0 */
	  if (inb_p(ser_io_base[port] + 5) & 0x01)
	    {
	      byte = inb_p(ser_io_base[port] + 0);
	      break;
	    }
	}


      ((char*)buffer)[i] = byte;
      bytes_read++;
    }

  return bytes_read;
}



/***********/
/*  Write  */
/***********/

ssize_t serial_port_write (int file_descriptor, void *buffer, size_t bytes)
{
  return basic_write (get_minor (file_descriptor), buffer, bytes);
}


/*****************/
/*  Basic Write  */
/*****************/

ssize_t basic_write (unsigned int port, void *buffer, size_t bytes)
{
  unsigned int i, bytes_written = 0;
  unsigned char mode;
  for(i=0; i<bytes; i++) {
    if (ser_termios[port].oflag & OPOST) {
      if (((char *)buffer)[i] == '\n') {
	/* Wait for the transmit buffer to become available.  */
	while (!(inb_p(ser_io_base[port] + 5) & 0x20));
	outb_p(ser_io_base[port] + 0, '\r');
      }
    }
    mode = inb_p(ser_io_base[port] + 3);
    outb_p(ser_io_base[port] + 3, 0x7F & mode);           /* DLAB = 0 */
    /* Wait for the transmit buffer to become available.  */
    while (!(inb_p(ser_io_base[port] + 5) & 0x20));
    outb_p(ser_io_base[port] + 0, ((char *)buffer)[i]);

    bytes_written++;
  }

  return bytes_written;
}


/***********/
/*  Ioctl  */
/***********/

int serial_port_ioctl (int file_descriptor, int request, void* argp)
{
  int port;
  unsigned char mode;
  unsigned freq_divisor;
  termios_t *com_params = (termios_t *) argp;

  port = get_minor (file_descriptor);

  switch (request)
    {
    case SETATTR:
      set_port_attr (port, com_params);
      break;


    case GETATTR:
      /* Clear struct data to get the parameters */
      com_params->iflag = 0x00000000;
      com_params->oflag = 0x00000000;
      com_params->cflag = 0x00000000;
      com_params->lflag = 0x00000000;
      com_params->ispeed = 0;
      com_params->ospeed = 0;

      mode = inb_p(ser_io_base[port] + 3);  /* reading the Line Control
					     * Register (LCR) */

      switch (mode & 0x03)      /* Length of the word  (WLS1 - WLS0) */
	{
	case 0x00: com_params->cflag |= CS5; break;
	case 0x01: com_params->cflag |= CS6; break;
	case 0x02: com_params->cflag |= CS7; break;
	case 0x03: com_params->cflag |= CS8; break;
	}

      if (mode & 0x08)    /* Parity bit enable  (PEN) */
	{
	  com_params->cflag |= PARENB;

	  if (!(mode & 0x10))    /* Odd parity (PEN) */
	    com_params->cflag |= PARODD;
	}


      if (mode & 0x04)    /* Stop bits number  (STB) */
	com_params->cflag |= CSTOPB;

      /* DLAB = 1 */
      outb_p(ser_io_base[port] + 3, 0x80 | mode);
      freq_divisor = 
	inb_p(ser_io_base[port] + 1) << 8 | inb_p(ser_io_base[port] + 0);
         /* MSB latch divisor */            /* LSB latch divisor */

      /* Convert the divisor latch value into a baud rate. */
      com_params->ispeed = 115200 / freq_divisor;
      com_params->ospeed = 115200 / freq_divisor;
      break;


    case SETSPEED:
      /* Convert the baud rate into a divisor latch value.  */
      freq_divisor = 115200 / com_params->ospeed;

      /* Initialize the serial port.  */
      mode = inb_p(ser_io_base[port] + 3);  /* reading the Line Control
					     * Register (LCR) */
      outb_p(ser_io_base[port] + 3, 0x80 | mode);           /* DLAB = 1 */
      outb_p(ser_io_base[port] + 0, freq_divisor & 0x00FF); /* LSB latch
							     * divisor */
      outb_p(ser_io_base[port] + 1, freq_divisor >> 8); /* MSB latch divisor */

      break;


    case GETSPEED:
      mode = inb_p(ser_io_base[port] + 3);  /* reading the Line Control
					     * Register (LCR) */
      outb_p(ser_io_base[port] + 3, 0x80 | mode);    /* DLAB = 1 */
      freq_divisor =
	inb_p(ser_io_base[port] + 1) << 8 | inb_p(ser_io_base[port] + 0);
         /* MSB latch divisor */            /* LSB latch divisor */

      /* Convert the divisor latch value into a baud rate. */
      com_params->ispeed = 115200 / freq_divisor;
      com_params->ospeed = 115200 / freq_divisor;
      break;


    case EINTERRUPT:
      outb_p(ser_io_base[port] + 1, 0x01);  /* interrupt on received data */
      break;


    case FLUSH:
      /*
       * Line Status Register (LSR)  :
       * bit #6 is 'HIGH' if there is no byte in hold register or shift
       * register
       */
      while (!(inb_p(ser_io_base[port] + 5) & 0x40));
      break;


    default:
      break;

    }

  return 0;
}

/*******************/
/*  set_port_attr  */
/*******************/
static void set_port_attr (int port, termios_t *com_params)
{
  unsigned char mode;
  unsigned freq_divisor;
  /* Determine what to plug in the data format register.  */
  if (com_params->cflag & PARENB)
    {
      if (com_params->cflag & PARODD)
	mode = 0x08;
      else
	mode = 0x18;
    }
  else
    mode = 0x00;

  if (com_params->cflag & CSTOPB)
    mode |= 0x04;

  switch (com_params->cflag & 0x00000300)
    {
    case CS5: mode |= 0x00; break;
    case CS6: mode |= 0x01; break;
    case CS7: mode |= 0x02; break;
    case CS8: mode |= 0x03; break;
    }

  /* Convert the baud rate into a divisor latch value.  */
  freq_divisor = 115200 / com_params->ospeed;

  /* Initialize the serial port.  */
  outb_p(ser_io_base[port] + 3, 0x80 | mode);           /* DLAB = 1 */
  outb_p(ser_io_base[port] + 0, freq_divisor & 0x00FF); /* LSB latch divisor */
  outb_p(ser_io_base[port] + 1, freq_divisor >> 8); /* MSB latch divisor */

  outb_p(ser_io_base[port] + 3, 0x7F & mode);    /* DLAB = 0 */
  outb_p(ser_io_base[port] + 1, 0x00);     /* no interrupts enabled */
  outb_p(ser_io_base[port] + 4, 0x0B);     /* OUT2, RTS, and DTR enabled */

  /* make sure the FIFO is on */
  outb_p(ser_io_base[port] + 2, 0x41); /* 4 byte trigger (0x40); on (0x01) */

  /* Clear all serial interrupts.  */
  inb_p(ser_io_base[port] + 6);   /* ID 0: read RS-232 status register */
  inb_p(ser_io_base[port] + 2);   /* ID 1: read interrupt identification reg */
  inb_p(ser_io_base[port] + 0);   /* ID 2: read receive buffer register */
  inb_p(ser_io_base[port] + 5);   /* ID 3: read serialization status reg */
}
